/* hcd -- solve the Highly Connected Deletion problem
   Copyright (C) 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  */

#ifndef C_GRAPH_INCLUDED
#define C_GRAPH_INCLUDED

#include <cassert>
#include <cstdint>
#include <climits>

#include <vector>
#include <unordered_map>
#include <iostream>

typedef int vertex_t;
typedef double weight_t;

class NeighborIt {
public:
    NeighborIt(std::unordered_map<vertex_t, weight_t>::const_iterator it) : it_(it) { }
    bool operator!=(const NeighborIt other) { return it_ != other.it_; }
    NeighborIt& operator++() { ++it_; return *this; }
    int n() const { return it_->first; }
    weight_t w() const { return it_->second; }
    
private:
    std::unordered_map<vertex_t, weight_t>::const_iterator it_;
};

class Graph {
public:
    Graph(int n = 0) : neighbors_(n) { }

    int num_vertices() const { return neighbors_.size(); }
    int num_edges() const {
	int n = 0;
	for (std::size_t i = 0; i < neighbors_.size(); i++)
	    n += neighbors_[i].size();
	return n / 2;
    }
    bool has_vertex(int u) const { return u >= 0 && u < num_vertices(); }
    int deg(int u) const {
	assert(has_vertex(u));
	return neighbors_[u].size();
    }
    bool is_connected(int u, int v) const {
	assert(has_vertex(u) && has_vertex(v));
	return neighbors_[u].find(v) != neighbors_[u].end();
    }
    weight_t weight(int u, int v) const {
	assert(has_vertex(u) && has_vertex(v));
	return neighbors_[u].find(v)->second;
    }
    NeighborIt neighbors_begin(int u) const {
	assert(has_vertex(u));
	return neighbors_[u].begin();
    }
    NeighborIt neighbors_end(int u) const {
	assert(has_vertex(u));
	return neighbors_[u].end();
    }

    void add_vertex() { neighbors_.push_back(std::unordered_map<int, weight_t>()); }
    void connect(int u, int v, weight_t w) {
	assert(has_vertex(u) && has_vertex(v));
	neighbors_[u][v] = w;
	neighbors_[v][u] = w;
    }
    void disconnect(int u, int v) {
	assert(is_connected(u, v));
	neighbors_[u].erase(neighbors_[u].find(v));
	neighbors_[v].erase(neighbors_[v].find(u));
    }
    void set_weight(int u, int v, weight_t w) { connect(u, v, w); }

private:
    std::vector<std::unordered_map<vertex_t, weight_t> > neighbors_;    
};

inline std::ostream& operator<<(std::ostream& out, const Graph& g) {
    for (int u = 0; u < g.num_vertices(); u++)
	if (g.deg(u) == 0)
	    out << u << std::endl;
    for (int u = 0; u < g.num_vertices(); u++)
	for (NeighborIt it = g.neighbors_begin(u); it != g.neighbors_end(u); ++it)
	    if (it.n() >= u)
		out << u << ' ' << it.n() << ' ' << it.w() << std::endl;
    return out;
}

struct Edge {
    Edge() { }
    Edge(int nu, int nv) {
	if (nu < nv) {
	    u = nu;
	    v = nv;
	} else {
	    u = nv;
	    v = nu;
	}
    }
    bool operator==(const Edge& other) const { return u == other.u && v == other.v; }
    int u, v;
};

template <class T>
inline void hash_combine(std::size_t& seed, const T& v) {
    std::hash<T> hasher;
    seed ^= hasher(v) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
}

namespace std {
    template<>
    struct hash<Edge> {
	size_t operator()(const Edge& e) const {
	    size_t seed = 0;
	    hash_combine(seed, e.u);
	    hash_combine(seed, e.v);
	    return seed;
	}
    };
}

inline bool sim(weight_t x, weight_t y) {
    return abs(x - y) < 1e-8;
}

#endif
