(* colcom -- solve the colorful components problem
   Copyright (C) 2011, 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

(*
  in:
  n_c: size of component
  n: number of vertices
  p_v: prob. component contains a vertex
  p_e: prob. component contains an edge between two vertices
  p_x: prob. unrelated vertices are connected
*)

let choose_2 x = (x * (x - 1)) / 2;;
let (//.) x y = (float_of_int x) /. (float_of_int y);;
let round x = int_of_float (floor (x +. 0.5));;
  
let fold_vertex_pairs f g accu =
  Graph.fold_vertices
    (fun accu u _ ->
      Graph.fold_vertices
        (fun accu v _ ->
          if u >= v then accu
          else f accu u v)
        g accu)
    g accu
;;

let generate n_c n p_v p_e p_x =
  let g = Graph.empty in
  let rec make_components g component i =
    if Graph.num_vertices g >= n
    then g, component
    else
      let g, component =
	Util.fold_n
	  (fun (g, component) c ->
	    if Random.float 1.0 <= p_v then
	      let g, v = Graph.new_vertex g in
	      let component = IntMap.add component v i in
	      g, component
	    else
	      g, component)
	  n_c (g, component)
      in make_components g component (i + 1)
  in
  let rec loop () =
    let g, components = make_components g IntMap.empty 0 in
    if Graph.num_vertices g = n
    then g, components
    else loop () in
  let g, components = loop () in
  let m_max = choose_2 n in
  
  let sizes = IntMap.fold (fun sizes _ i -> IntMap.modify_default succ sizes i 0) components IntMap.empty in
  let m_internal_max = IntMap.fold (fun m_internal_max _ size -> m_internal_max + choose_2 size) sizes 0 in
  
  let m_external_max = m_max - m_internal_max in
  let m_internal = round ((float_of_int m_internal_max) *. p_e) in
  let m_external = round ((float_of_int m_external_max) *. p_x) in
  let rec loop g m_internal m_external =
    if m_internal = 0 && m_external = 0
    then g
    else
      let u = Random.int n in
      let v = Random.int n in
      if u = v || Graph.is_connected g u v then loop g m_internal m_external
      else
	if m_internal > 0 && IntMap.get components u = IntMap.get components v
	then loop (Graph.connect g u v) (m_internal - 1) m_external
	else if m_external > 0 && IntMap.get components u <> IntMap.get components v
	then loop (Graph.connect g u v) m_internal (m_external - 1)
	else loop g m_internal m_external
  in
  loop g m_internal m_external, components
;;

let main () =
  if Array.length Sys.argv >= 7
  then Random.init (int_of_string Sys.argv.(6))
  else Random.self_init ();
  let n   = int_of_string   Sys.argv.(1) in
  let n_c = int_of_string   Sys.argv.(2) in
  let p_v = float_of_string Sys.argv.(3) in
  let p_e = float_of_string Sys.argv.(4) in
  let p_x = float_of_string Sys.argv.(5) in
  let g, components = generate n_c n p_v p_e p_x in
  Printf.printf "# n = %d n_c = %d p_v = %s p_e = %s p_x = %s\n"
    n n_c (string_of_float p_v) (string_of_float p_e) (string_of_float p_x);
  let clusters =
    IntMap.fold
      (fun clusters i c ->
	IntMap.modify_default (fun s -> IntSet.add s i) clusters c IntSet.empty)
      components IntMap.empty in
  IntMap.iter
    (fun _ vs ->
      let g' = Graph.subgraph g vs in
      Printf.printf "# n = %d m = %d:" (Graph.num_vertices g') (Graph.num_edges g');
      IntSet.iter (fun u -> Printf.printf " %d" u) vs;
      print_newline ())
    clusters;
  Graph.iter_edges (fun u v -> Printf.printf "%d\t%d\n" u v) g;
;;

main ();;
