(* hcd -- solve the Highly Connected Deletion problem
   Copyright (C) 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

type t = IntSet.t IntMap.t;;

let empty = IntMap.empty;;


(* Helper functions.  *)

let has_vertex g u = IntMap.has_key g u;;

let neighbors g u = IntMap.get g u;;

let deg g u = IntSet.size (neighbors g u);;

let fold_neighbors f g u z =
  assert (has_vertex g u);
  IntSet.fold f (neighbors g u) z
;;

let fold_vertices f g accu = IntMap.fold f g accu;;
let iter_vertices f g      = IntMap.iter f g;;

let fold_vertex_pairs f g accu =
  IntMap.fold
    (fun accu u n_u ->
      IntMap.fold
	(fun accu v n_v ->
	  if u < v then f accu u n_u v n_v else accu)
	g accu)
    g accu
;;

let fold_edges f g accu =
  fold_vertices
    (fun accu u neighbors ->
      IntSet.fold
        (fun accu v -> if u < v then f accu u v else accu)
        neighbors
        accu)
    g accu
;;

let iter_edges f g = fold_edges (fun () u v -> f u v) g ();;


(* Querying graphs.  *)

let num_vertices g = IntMap.size g;;
let vertex_set g = fold_vertices (fun vs u _ -> IntSet.add vs u) g IntSet.empty;;

let num_edges g =
  let num = fold_vertices (fun num u n_u -> num + IntSet.size n_u) g 0 in
  assert (num mod 2 = 0);
  num / 2
;;

let is_connected g u v = IntSet.contains (neighbors g u) v;;

let fold_connected_components f g z =
  let rec dfs seen component u =
    fold_neighbors
      (fun (seen, component) v ->
	if IntSet.contains seen v
	then seen, component
	else
	  let seen = IntSet.add seen v in
	  let component = IntMap.add component v () in
	  dfs seen component v)
      g u (seen, component)
  in
  fst
    (fold_vertices
       (fun (z, seen) u _ ->
	 if IntSet.contains seen u
	 then (z, seen)
	 else
	   let seen, component = dfs (IntSet.add seen u) (IntMap.singleton u ()) u in
	   f z (IntMap.intersection g component), seen)
       g (z, IntSet.empty))
;;


(* Modifying graphs.  *)

let add_vertex g u = IntMap.add g u IntSet.empty;;
let new_vertex g =
  let i = if IntMap.is_empty g then 0 else IntMap.max_key g + 1 in
  add_vertex g i, i
;;

let delete_vertex g u =
  IntSet.fold
    (fun g v -> IntMap.modify (fun n_v -> IntSet.delete n_v u) g v)
    (neighbors g u)
    (IntMap.delete g u)
;;

let connect g u v =
  if u = v then invalid_arg "Graph.connect: self-loop added";
  let g = IntMap.modify (fun n_u -> IntSet.add n_u v) g u in
  let g = IntMap.modify (fun n_v -> IntSet.add n_v u) g v in
  g
;;

let disconnect g u v =
  let g = IntMap.modify (fun n_u -> IntSet.delete n_u v) g u in
  let g = IntMap.modify (fun n_v -> IntSet.delete n_v u) g v in
  g
;;

let subgraph g vs =
  IntSet.fold
    (fun g' u -> IntMap.add g' u (IntSet.intersection (neighbors g u) vs))
    vs empty
;;


(* I/O  *)

let output channel g =
  (* Printf.fprintf channel "{ n = %d, m = %d\n" (num_vertices g) (num_edges g); *)
  iter_vertices (fun u n_u ->
                   if IntSet.is_empty n_u
                   then Printf.fprintf channel "%d\n" u) g;
  iter_edges (fun u v -> Printf.fprintf channel "%d %d\n" u v) g;
  (* Printf.fprintf channel "}\n"; *)
;;

(* try int_of_string... is not what we want, since it accepts e.g. "0x17"  *)
let is_int s =
  let isdigit = function
    | '0' .. '9' -> true
    | _ -> false in
  let rec loop p =
    if p >= String.length s
    then true
    else isdigit s.[p] && loop (p + 1)
  in
  String.length s > 0 && loop 0
;;

module StringMap = Map.Make(String);;

let input channel =
  let g              = ref empty in
  let vertex_numbers = ref StringMap.empty in
  let vertex_names   = ref IntMap.empty in
  let lineno         = ref 0 in
  let warned         = ref 0 in
  let next_key nu =
    if is_int nu && not (IntMap.has_key !vertex_names (int_of_string nu)) then int_of_string nu
    else if IntMap.is_empty !vertex_names then 0
    else (IntMap.max_key !vertex_names) + 1 in
  let parse_vertex nu =
    if StringMap.mem nu !vertex_numbers then StringMap.find nu !vertex_numbers
    else
      let u = next_key nu in
      vertex_numbers := StringMap.add nu u !vertex_numbers;
      vertex_names   := IntMap.add !vertex_names u nu;
      g := add_vertex !g u;
      u in
  let parse_edge nu nv =
    let u = parse_vertex nu in
    let v = parse_vertex nv in
    if u = v then begin
      Printf.eprintf "line %d: self-loop for vertex '%s'\n" !lineno nu;
      exit 1;
    end;
    if is_connected !g u v
    then begin
      (match !warned with
      | 0 -> Printf.eprintf "warning: line %d: edge (%s, %s) already exists; ignoring\n%!" !lineno nu nv;
      | 1 -> Printf.eprintf "warning: suppressing further warnings\n%!";
      | _ -> ());
      incr warned;
    end
    else g := connect !g u v;
  in
  (try
     while true do
       let line = input_line channel in
       incr lineno;
       let line =
         if String.contains line '#'
         then String.sub line 0 (String.index line '#')
         else line in
       let line = Util.split_string line in
       match line with
       | [] -> ()
       | [u; v] -> parse_edge u v
       | _ -> begin
	 Printf.eprintf "line %d: syntax error\n" !lineno;
	 exit 1;
       end;
     done;
   with End_of_file -> ());
  (!g, !vertex_names, !vertex_numbers)
;;
