(* hcd -- solve the Highly Connected Deletion problem
   Copyright (C) 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

(** Undirected functional graphs. Vertices are represented by
    nonnegative integers.  No multiple edges or self-loops are
    possible.  *)

(** The type of a graph.  *)
type t

(** The empty graph.  *)
val empty : t


(** Querying graphs.  *)

(** Number of vertices. O(1) time.  *)
val num_vertices : t -> int

(** Returns the degree of a vertex. O(n log n) time.  *)
val deg : t -> int -> int

(** Returs the set of vertices. O(n log n) time.  *)
val vertex_set : t -> IntSet.t

(** [neighbors g u] returns the set of neighbors of [u] in [g].  *)
val neighbors : t -> int -> IntSet.t

(** Number of edges. O(n) time.  *)
val num_edges : t -> int

(** Returns true if two vertices are connected.  O(log n) time.  *)
val is_connected : t -> int -> int -> bool

(** [fold_vertices f g a] computes [(f uN nN ... (f u1 n1 a)...)],
    where [u1, ... iN] are the vertices of [g] and [nN] is the set of
    neighbors of [uN]. O(n) time.  *)
val fold_vertices :
  ('a -> int -> IntSet.t -> 'a) -> t -> 'a -> 'a

(** [iter_vertices f g] calls [f u nu] for each vertex [u] in [g],
    where [nu] is the set of neighbors of [u].  *)
val iter_vertices :
  (int -> IntSet.t -> unit) -> t -> unit

(** [fold_neighbors f g u a] computes [(f vN ... (f v1 a)...)],
    where [v1, ..., vN] are the neighbors of [u] in [g].
    O(log n + deg u) time.  *)
val fold_neighbors :
  ('a -> int -> 'a) -> t -> int -> 'a -> 'a

(** [fold_vertices f g a] computes [(f uN nN vN mN ... (f u1 n1 v1 m1
    a)...)], where [u1, ... uN] and [v1, ... vN] are the vertices of
    [g] with [i] < [j], [nN] is the neighbor set of [uN], and [mN] is
    the neighbor set of [vN].  O(n^2) time.  *)
val fold_vertex_pairs :
  ('a -> int -> IntSet.t -> int -> IntSet.t -> 'a) -> t -> 'a -> 'a

(** [fold_edges f g a] computes [(f uN vN ... (f u1 v1 a)...)], where
    [(u1, v1) ... (uN, vN)] are the edges of [g].  O(m) time.  *)
val fold_edges : ('a -> int -> int -> 'a) -> t -> 'a -> 'a

(** [iter_edges f g a] calls [f i1 j1, ..., f iN jN] where [(i1, j1)
    ... (iN, jN)] are the edges of [g].  O(m) time.  *)
val iter_edges : (int -> int -> unit) -> t -> unit

val fold_connected_components : ('a -> t -> 'a) -> t -> 'a -> 'a


(** Modifying graphs.  *)

(** [add_vertex g i returns [g] with an additional new vertex [i].
    Raises [IntMap.Already_present] if [i] is already a vertex.  O(log
    n) time.  *)
val add_vertex : t -> int -> t

(** [new_vertex g] returns [g] with an additional new vertex [i],
    and the new vertex [i].  O(log n) time.  *)
val new_vertex : t -> t * int

(** [delete_vertex g] returns [g] without vertex [u]. Raises
    [Not_found] if [u] is not in [g]. O(log n * deg u) time.  *)
val delete_vertex : t -> int -> t

(** [connect g u v] returns [g] with vertices [u] and [v] connected.
    Raises [Not_found] when [u] or [v] do not exist in [g]. Raises
    [IntSet.Already_present] when [u] and [v] are already connected.
    Raises [Invalid_argument] when [u = v].  O(log n) time.  *)
val connect : t -> int -> int -> t

val disconnect : t -> int -> int -> t

(** Subgraph induced by a vertex set.  *)
val subgraph : t -> IntSet.t -> t


(** I/O  *)

module StringMap :
  sig
    type key = string
    type 'a t = 'a Map.Make(String).t
    val empty : 'a t
    val is_empty : 'a t -> bool
    val add : key -> 'a -> 'a t -> 'a t
    val find : key -> 'a t -> 'a
    val remove : key -> 'a t -> 'a t
    val mem : key -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val map : ('a -> 'b) -> 'a t -> 'b t
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val compare : ('a -> 'a -> int) -> 'a t -> 'a t -> int
    val equal : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool
  end

(** [output c g] prints a debug representation of [g] to channel [c].
    O(m) time.  *)
val output : out_channel -> t -> unit

val input : in_channel -> t * string IntMap.t * int StringMap.t
  
