(* hcd -- solve the Highly Connected Deletion problem
   Copyright (C) 2011, 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

let current_g = ref Graph.empty;;

(* minimizing weight! *)
let colgen_heuristic weights =
  let g = !current_g in
  let g_x = Graph.fold_vertex_pairs
    (fun g_x u n_u v n_v ->
      if Graph.is_connected g u v || IntSet.intersection_size n_u n_v >= 3
      then Graph.connect g_x u v
      else g_x)
    g (Graph.fold_vertices (fun g_x u _ -> Graph.add_vertex g_x u) g Graph.empty) in
  let is_hc vs =
    let n_half = IntSet.size vs / 2 in
    IntSet.fold (fun yes u -> yes && IntSet.intersection_size (Graph.neighbors g u) vs > n_half) vs true in
  let vs, vs_val =
    Graph.fold_vertices
      (fun (best, best_val) u n_u ->
	(* Printf.eprintf "u = %d best = %a best_val = %f\n%!" u IntSet.output best best_val; *)
	let rec loop best best_val vs vs_val n_vs n_vs_x =
	  (* Printf.eprintf "vs = %a val = %f n_vs = %a\n%!" IntSet.output vs vs_val IntSet.output n_vs; *)
	  let best, best_val = if is_hc vs && vs_val < best_val
	    then vs, vs_val
	    else best, best_val in
	  if IntSet.is_empty n_vs then best, best_val
	  else
	    let v, v_val = IntSet.fold
	      (fun (best_v, best_v_val) v ->
		let v_val =
		  weights.(v)
		  -. float_of_int (IntSet.intersection_size (Graph.neighbors g v) vs)
		  (* -. 1e-6 *. float_of_int (IntSet.intersection_size (Graph.neighbors g_x v) n_vs_x) *)
		in
		if v_val < best_v_val
		then v, v_val
		else best_v, best_v_val)
	      n_vs (-1, infinity) in
	    let vs = IntSet.add vs v in
	    let vs_val = vs_val +. v_val in
	    let n_vs_x = IntSet.intersection n_vs_x (Graph.neighbors g_x v) in
	    let n_vs = IntSet.union n_vs (Graph.neighbors g v) in
	    let n_vs = IntSet.intersection n_vs n_vs_x in
	    loop best best_val vs vs_val n_vs n_vs_x in
	loop best best_val (IntSet.singleton u) weights.(u) (Graph.neighbors g u) (Graph.neighbors g_x u))
      g (IntSet.empty, infinity)
  in
  IntSet.to_array vs, vs_val
;;

let independent_set_heuristic weights =
  let g = !current_g in
  let g_x = Graph.fold_vertex_pairs
    (fun g_x u n_u v n_v ->
      if Graph.is_connected g u v || IntSet.intersection_size n_u n_v >= 3
      then Graph.connect g_x u v
      else g_x)
    g (Graph.fold_vertices (fun g_x u _ -> Graph.add_vertex g_x u) g Graph.empty) in
  let weights = Array.mapi (fun i w -> w /. float_of_int (Graph.deg g_x i)) weights in
  let rec loop is cand =
    if IntSet.is_empty cand
    then is
    else
      let best, _ = IntSet.fold
	(fun (best, w_best) u ->
	  let w_u = weights.(u) in if w_u > w_best then u, w_u else best, w_best) cand (-1, -.infinity) in
      let cand = IntSet.delete cand best in
      let cand = IntSet.minus cand (Graph.neighbors g_x best) in
      loop (IntSet.add is best) cand in
  let is = loop IntSet.empty (Graph.vertex_set g_x) in
  let value = IntSet.fold (fun value u -> value +. weights.(u)) is 0.0 in
  IntSet.to_array is, value
;;

let () = Callback.register "colgen_heuristic" colgen_heuristic;;
let () = Callback.register "independent_set_heuristic" independent_set_heuristic;;
