/* colcom -- solve the colorful components problem
   Copyright (C) 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  */

#include <math.h>

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <limits>

#include "ILPCplex.hh"


static void cplex_error(CPXCENVptr env, int status, const char* msg) {
    char buffer[CPXMESSAGEBUFSIZE];
    fprintf(stderr, "error: %s: %s\n", msg, CPXgeterrorstring(env, status, buffer));
    exit(1);
}

ILPCplex::ILPCplex() {
    int status;
    // suppress annoying "license manager" output
    int backup, redirection;
    fflush(stderr);
    backup = dup(2);
    redirection = open("/dev/null", O_WRONLY);
    dup2(redirection, 2);
    close(redirection); 
    env_ = CPXopenCPLEX(&status);
    fflush(stderr);
    dup2(backup, 2);
    close(backup);
    if (!env_)
	error(status, "failed to create CPLEX environment");

    lp_ = CPXcreateprob(env_, &status, "colcom");
    if (!lp_)
	error(status, "failed to create CPLEX LP");
}

ILPCplex::~ILPCplex() {
    int status = CPXcloseCPLEX(&env_);
    if (status)
	error(status, "CPXcloseCPLEX");
}

int ILPCplex::numVars() const {
    return CPXgetnumcols(env_, lp_);
}

void ILPCplex::setVerbose(bool verbose) {
    int status;
    if ((status = CPXsetintparam(env_, CPX_PARAM_SCRIND, verbose ? CPX_ON : CPX_OFF)))
	error(status, "CPXsetintparam(CPX_PARAM_SCRIND)");
}

void ILPCplex::addVariables(int n, const double* obj, const double* lb, const double* ub,
			    const ILPVarType* type, char** name) {
    char cplex_type[n];
    for (int i = 0; i < n; i++) {
	switch (type[i]) {
	//case VAR_CONT: cplex_type[i] = 'C'; break;
	case VAR_BIN:  cplex_type[i] = 'B'; break;
	case VAR_INT:  cplex_type[i] = 'I'; break;
	default: error(0, "invalid variable type"); exit(1); break;
	}
    }
    int status = CPXnewcols(env_, lp_, n, obj, lb, ub, cplex_type, name);
    if (status)
	error(status, "CPXnewcols");
}

static char cplex_sense(ILPConstrSense sense) {
    switch (sense) {
    case CONSTR_LEQ: return 'L'; break;
    case CONSTR_EQ:  return 'E'; break;
    case CONSTR_GEQ: return 'G'; break;
    }
    cplex_error(NULL, 0, "invalid constraint sense");
    exit(1);
}

void ILPCplex::addConstraints(int n, int nz, const double* rhs, const ILPConstrSense* sense,
			      const int* rmatbeg, const int* rmatind, const double* rmatval) {
    char cplex_senses[n];
    for (int i = 0; i < n; i++) 
	cplex_senses[i] = cplex_sense(sense[i]);
    int status = CPXaddrows(env_, lp_, 0, n, nz, rhs, cplex_senses, rmatbeg, rmatind, rmatval, NULL, NULL);
    if (status)
	error(status, "CPXaddrows");
}

void ILPCplex::addMIPstart(const double* x) {
    int n = numVars();
    int mcnt = 1;
    int ms_beg[1] = {0};
    int ms_indices[n];
    for (int i = 0; i < n; i++)
	ms_indices[i] = i;
    int status = CPXaddmipstarts(env_, lp_, mcnt, n, ms_beg, ms_indices, x, NULL, NULL);
    if (status)
	error(status, "CPXaddmipstarts");
}

double CallbackILPCplex::gap() const {
  double gap;
  int status = CPXgetcallbackinfo(env_, cbdata_, wherefrom_, CPX_CALLBACK_INFO_MIP_REL_GAP, &gap);
  if (status)
    cplex_error(env_, status, "CPX_CALLBACK_INFO_MIP_REL_GAP");
  return gap >= CPX_INFBOUND ? std::numeric_limits<double>::infinity() : gap;
}

void CallbackILPCplex::addLazyConstraint(int nz, double rhs, ILPConstrSense sense,
					 const int* cutind, const double* cutval) {
    int status = CPXcutcallbackadd(env_, cbdata_, wherefrom_, nz, rhs, cplex_sense(sense),
				   cutind, cutval, CPX_USECUT_FORCE);
    if (status)
	cplex_error(env_, status, "CPXcutcallbackadd");
}

void CallbackILPCplex::addCut(int nz, double rhs, ILPConstrSense sense, const int* cutind, const double* cutval) {
    int status = CPXcutcallbackadd(env_, cbdata_, wherefrom_, nz, rhs, cplex_sense(sense),
				   cutind, cutval, CPX_USECUT_FORCE); // FIXME maybe others better?
    if (status)
	cplex_error(env_, status, "CPXcutcallbackadd");
}

static int cplex_callback(CPXCENVptr env, void* cbdata, int wherefrom, void* cbhandle, int* useraction_p) {
    //fprintf(stderr, "cplex_callback\n");
    ILPCallback& callback = *reinterpret_cast<ILPCallback*>(cbhandle);
    CallbackILPCplex ilp(env, cbdata, wherefrom);
    int status;

    CPXLPptr lp;
    if ((status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_USER_PROBLEM, &lp)))
	cplex_error(env, status, "CPX_CALLBACK_INFO_USER_PROBLEM");
    int n = CPXgetnumcols(env, lp);

    double x[n];
    if ((status = CPXgetcallbacknodex(env, cbdata, wherefrom, x, 0, n - 1)))
	cplex_error(env, status, "CPXgetcallbacknodex");

    bool isIntegral;
    switch (wherefrom) {
    case CPX_CALLBACK_MIP_CUT_LOOP:
    case CPX_CALLBACK_MIP_CUT_LAST:
	isIntegral = false; break;
    case CPX_CALLBACK_MIP_CUT_FEAS:
	isIntegral = true;
	// Scrub solution. Some values are sometimes not actually integral.
	// FIXME: some variables might not be integers.
	for (int i = 0; i < n; i++)
	    x[i] = round(x[i]);
	break;
    default:
	fprintf(stderr, "error: unknown wherefrom %d\n", wherefrom);
	exit(1);
    }

    callback.call(ilp, x, isIntegral);

    *useraction_p = CPX_CALLBACK_DEFAULT;
    return 0;
}

void ILPCplex::addCallback(ILPCallback& callback, bool cuts, bool lazyConstraints) {
    int status;
    // the callback wants to use variable indices of the original and not the presolved ILP
    if ((status = CPXsetintparam(env_, CPX_PARAM_MIPCBREDLP, CPX_OFF)))
	cplex_error(env_, status, "CPXsetintparam(CPX_PARAM_MIPCBREDLP)");
    // manual says to switch this off, too
    if ((status = CPXsetintparam(env_, CPX_PARAM_PRELINEAR, CPX_OFF)))
	cplex_error(env_, status, "CPXsetintparam(CPX_PARAM_PRELINEAR)");
    if (cuts) {
	if ((status = CPXsetusercutcallbackfunc(env_, cplex_callback, &callback)))
	    error(status, "CPXsetusercutcallbackfunc");
    }
    if (lazyConstraints) {
	if ((status = CPXsetlazyconstraintcallbackfunc(env_, cplex_callback, &callback)))
	    error(status, "CPXsetlazyconstraintcallbackfunc");
    }
}

double ILPCplex::optimize(double* x) {
    int status;

    //CPXwriteprob(env_, lp_, "/dev/stderr", "lp");

    // CPLEX errors out if there are 0 variables, so this needs to be special-cased
    if (CPXgetnumint(env_, lp_) + CPXgetnumbin(env_, lp_) == 0) {
	if ((status = CPXlpopt(env_, lp_)))
	    error(status, "CPXlpopt");
    } else {
	if ((status = CPXmipopt(env_, lp_)))
	    error(status, "CPXmipopt");
    }
    
    int lpstat;
    double objval;
    if ((status = CPXsolution(env_, lp_, &lpstat, &objval, x, NULL, NULL, NULL)))
	error(status, "CPXsolution");
    if (lpstat != CPXMIP_OPTIMAL && lpstat != CPX_STAT_OPTIMAL) {
	fprintf(stderr, "error: unknown CPLEX status %d\n", lpstat);
	exit(1);
    }
    return objval;
}

void ILPCplex::error(int status, const char* msg) {
    cplex_error(env_, status, msg);
    exit(1);
}
