/* colcom -- solve the colorful components problem
   Copyright (C) 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  */

#include <math.h>
#include <stdlib.h>

#include <limits>

#include "ILPGurobi.hh"


void gurobi_error(GRBenv* env, const char* msg) {
    fprintf(stderr, "error: %s: %s\n", msg, GRBgeterrormsg(env));
    exit(1);
}

ILPGurobi::ILPGurobi() : cuts_(false), lazyConstraints_(false) {
    if (GRBloadenv(&env_, NULL))
	error("GRBloadenv");
    if (GRBnewmodel(env_, &lp_, "colcom", 0, NULL, NULL, NULL, NULL, NULL))
	error("GRBnewmodel");

    // each model gets its own copy of the environment when it is created
    orig_env_ = env_;
    env_ = GRBgetenv(lp_);

    setVerbose(false);

    // TODO:
    // Parameters to try:
    // CliqueCuts
    // MIPFocus
    // PreDepRow
    // PreSparsify
    // Presolve
    // Symmetry
    // Threads
    // VarBranch
}

ILPGurobi::~ILPGurobi() {
    if (GRBfreemodel(lp_))
	error("GRBfreemodel");
    GRBfreeenv(orig_env_);
}

int ILPGurobi::numVars() const {
    int n;
    if (GRBgetintattr(lp_, "NumVars", &n))
	error("GRBgetintattr(NumVars)");
    return n;
}

void ILPGurobi::setVerbose(bool verbose) {
    if (GRBsetintparam(env_, "OutputFlag", verbose))
	error("GRBsetintparam(OutputFlag)");
}

void ILPGurobi::addVariables(int n, const double* obj, const double* lb, const double* ub,
			    const ILPVarType* type, char** name) {
    char gurobi_type[n];
    for (int i = 0; i < n; i++) {
	switch (type[i]) {
	//case VAR_CONT: gurobi_type[i] = GRB_CONTINUOUS; break;
	case VAR_BIN:  gurobi_type[i] = GRB_BINARY; break;
	case VAR_INT:  gurobi_type[i] = GRB_INTEGER; break;
	default: error("invalid variable type"); break;
	}
    }

    if (GRBaddvars(lp_, n, 0, NULL, NULL, NULL,
		   const_cast<double*>(obj),
		   const_cast<double*>(lb),
		   const_cast<double*>(ub),
		   gurobi_type, name))
	error("GRBaddvars");
}

static char gurobi_sense(ILPConstrSense sense) {
    switch (sense) {
    case CONSTR_LEQ: return GRB_LESS_EQUAL; break;
    case CONSTR_EQ:  return GRB_EQUAL; break;
    case CONSTR_GEQ: return GRB_GREATER_EQUAL; break;
    }
    fprintf(stderr, "invalid constraint sense %d\n", sense);
    exit(1);
}

void ILPGurobi::addConstraints(int n, int nz, const double* rhs, const ILPConstrSense* sense,
			      const int* rmatbeg, const int* rmatind, const double* rmatval) {
    char gurobi_senses[n];
    for (int i = 0; i < n; i++) 
	gurobi_senses[i] = gurobi_sense(sense[i]);

    if (GRBupdatemodel(lp_))
	error("GRBupdatemodel");
    if (GRBaddconstrs(lp_, n, nz,
		      const_cast<int*>(rmatbeg),
		      const_cast<int*>(rmatind),
		      const_cast<double*>(rmatval),
		      gurobi_senses,
		      const_cast<double*>(rhs),
		      NULL))
	error("GRBaddconstrs");
    if (GRBupdatemodel(lp_))
	error("GRBupdatemodel");
}

void ILPGurobi::addMIPstart(const double* x) {
    int n = numVars();
    if (GRBsetdblattrarray(lp_, "Start", 0, n, const_cast<double*>(x)))
	error("GRBsetdblattrarray(Start)");
}

class CallbackILPGurobi : public CallbackILP {
public:
    CallbackILPGurobi(GRBenv* env, void* cbdata, int where)
	: env_(env), cbdata_(cbdata), where_(where) { }
    double gap() const {
	double best, bound;
	if (GRBcbget(cbdata_, where_, where_ == GRB_CB_MIPSOL ? GRB_CB_MIPSOL_OBJBST : GRB_CB_MIP_OBJBST, &best))
	    gurobi_error(env_, "GRBcbget(GRB_CB_MIP_OBJBST)");
	if (GRBcbget(cbdata_, where_, where_ == GRB_CB_MIPSOL ? GRB_CB_MIPSOL_OBJBND : GRB_CB_MIP_OBJBND, &bound))
	    gurobi_error(env_, "GRBcbget(GRB_CB_MIP_OBJBND)");
	double gap = fabs(bound - best) / best;
	return gap; // FIXME >= CPX_INFBOUND ? std::numeric_limits<double>::infinity() : gap;
    }
    void addLazyConstraint(int nz, double rhs, ILPConstrSense sense, const int* cutind, const double* cutval) {
	if (GRBcblazy(cbdata_, nz, cutind, cutval, gurobi_sense(sense), rhs))
	    gurobi_error(env_, "GRBcblazy");
    }
    void addCut(int nz, double rhs, ILPConstrSense sense, const int* cutind, const double* cutval) {
	if (GRBcbcut(cbdata_, nz, cutind, cutval, gurobi_sense(sense), rhs))
	    gurobi_error(env_, "GRBcbcut");
    }

private:
    GRBenv* env_;
    void* cbdata_;
    int where_;
};

static int gurobi_callback(GRBmodel *lp, void *cbdata, int where, void *usrdata) {
    if (where != GRB_CB_MIPNODE && where != GRB_CB_MIPSOL)
	return 0;

    ILPGurobi& data = *reinterpret_cast<ILPGurobi*>(usrdata);
    CallbackILPGurobi ilp(GRBgetenv(lp), cbdata, where);

    double x[data.numVars()];
    if (where == GRB_CB_MIPNODE && data.useCuts()) {
	int status;
	if (GRBcbget(cbdata, where, GRB_CB_MIPNODE_STATUS, &status))
	    gurobi_error(GRBgetenv(lp), "GRBcbget(GRB_CB_MIPNODE_STATUS)");
	if (status == GRB_LOADED || status == GRB_INFEASIBLE || status == GRB_CUTOFF)
	    return 0;
	if (status != GRB_OPTIMAL) {
	    fprintf(stderr, "unknown Gurobi status %d\n", status);
	    exit(1);
	}
	if (GRBcbget(cbdata, where, GRB_CB_MIPNODE_REL, x))
	    gurobi_error(GRBgetenv(lp), "GRBcbget(GRB_CB_MIPNODE_REL)");
	data.getCallback()->call(ilp, x, false);
    } else if (where == GRB_CB_MIPSOL && data.useLazyConstraints()) {
	if (GRBcbget(cbdata, where, GRB_CB_MIPSOL_SOL, x))
	    gurobi_error(GRBgetenv(lp), "GRBcbget(GRB_CB_MIPSOL_SOL)");
	// Scrub solution. Some values are sometimes not actually integral.
	// FIXME: some variables might not be integers.
	for (int i = 0; i < data.numVars(); i++)
	    x[i] = round(x[i]);	
	data.getCallback()->call(ilp, x, true);
    }

    return 0;
}

void ILPGurobi::addCallback(ILPCallback& callback, bool cuts, bool lazyConstraints) {
    cuts_ = cuts;
    lazyConstraints_ = lazyConstraints;
    callback_ = &callback;
    // When adding your own cuts, you must set parameter PreCrush to
    // value 1. This setting shuts off a few presolve reductions that
    // sometimes prevent cuts on the original model from being applied
    // to the presolved model.
    if (GRBsetintparam(env_, "PreCrush", 1))
	error("GRBsetintparam(PreCrush)");

    if (lazyConstraints) {
	// Turn off dual reductions - required when using lazy constraints 
	if (GRBsetintparam(env_, "DualReductions", 0))
	    error("GRBsetintparam(DualReductions)");
    }

    if (GRBsetcallbackfunc(lp_, gurobi_callback, this))
	error("GRBsetcallbackfunc");
}

double ILPGurobi::optimize(double* x) {
    if (GRBoptimize(lp_))
	error("GRBoptimize");

    if (GRBgetdblattrarray(lp_, "X", 0, numVars(), x))
	error("GRBgetdblattrarray(X)");

    int status;
    if (GRBgetintattr(lp_, "Status", &status))
	error("GRBgetintattr(Status)");
    if (status != GRB_OPTIMAL) {
	fprintf(stderr, "error: unknown Gurobi status %d\n", status);
	exit(1);
    }

    double objval;
    if (GRBgetdblattr(lp_, "ObjVal", &objval))
	error("GRBgetdblattr(ObjVal)");

    return objval;
}

void ILPGurobi::error(const char* msg) const {
    gurobi_error(env_, msg);
}
