(* hcd -- solve the Highly Connected Deletion problem
   Copyright (C) 2011, 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

(** The map is represented as big-endian patricia tree.  *)
type 'a t =
    Empty
  | Leaf of int * 'a
  | Branch of int * int * int * 'a t * 'a t
    (** [Branch (p, m, c, l, r)]:
        [p] is the largest common prefix for all the keys in this tree
        [m] is the branching bit mask
            ([m] is a power of 2, only the bits above [m] are valid in [p])
        [c] is the number of leaves of the tree (number of elements)
        [l] contains all the keys with a 0 in the branching bit
        [r] contains all the keys with a 1 in the branching bit  *)
;;

exception Already_present;;

let empty = Empty;;
let singleton i x = Leaf (i, x);;

let is_empty s = (s = Empty);;

let size = function
    Empty -> 0
  | Leaf _ -> 1
  | Branch (_, _, c, _, _) -> c
;;

let rec max_key = function
    Empty -> raise Not_found
  | Leaf (i, _) -> i
  | Branch (_, _, _, _, r) -> max_key r
;;

let rec min_key = function
    Empty -> raise Not_found
  | Leaf (i, _) -> i
  | Branch (_, _, _, l, _) -> min_key l
;;

let rec has_key s i =
  if i < 0 then invalid_arg "IntMap.has_key: negative key";
  match s with
      Empty -> false
    | Leaf (j, _) -> j = i
    | Branch (p, _, _, l, r) ->
	if i <= p
	then has_key l i
	else has_key r i
;;

let rec get s i =
  if i < 0 then invalid_arg "IntMap.get: negative key";
  match s with
      Leaf (j, x) when j = i -> x
    | Branch (p, _, _, l, r) ->
	if i <= p
	then get l i
	else get r i
    | _  -> raise Not_found
;;

let rec find s i =
  if i < 0 then invalid_arg "IntMap.get: negative key";
  match s with
      Leaf (j, x) when j = i -> Some x
    | Branch (p, _, _, l, r) ->
	if i <= p
	then find l i
	else find r i
    | _  -> None
;;

let rec get_default s i x =
  if i < 0 then invalid_arg "IntMap.get_default: negative key";
  match s with
      Empty -> x
    | Leaf (j, x) when j = i -> x
    | Leaf _ -> x
    | Branch (p, _, _, l, r) ->
	if i <= p
	then get_default l i x
	else get_default r i x
;;

let rec fold f s accu = match s with
    Empty -> accu
  | Leaf (i, x) -> f accu i x
  | Branch (_, _, _, l, r) -> fold f r (fold f l accu)
;;

let iter f s = fold (fun () k x -> f k x) s ();;

let rec any f = function
    Empty -> false
  | Leaf (i, x) -> f i x
  | Branch (_, _, _, l, r) -> any f l || any f r
;;


(* Modifying maps.  *)

(* Return an integer where only the highest bit that was set in [x] is
   still set.  *)
let branching_bit m p1 p2 =
  let rec highest_bit x =
    let x' = x land (x - 1) in
      if x' = 0
      then x
      else highest_bit x' in
  let x = p1 lxor p2 in			(* remove common prefix *)
  let x = x land lnot (m - 1) in	(* remove invalid suffix *)
    highest_bit x
;;

(* In [i], clear the 1-bit mask [m], and set all bits below [m]'s bit to one.  *)
let mask i m = (i lor (m - 1 + m)) - m;;

let prefix_matches i p m = (mask i m = p);;

let join m p1 s1 p2 s2 c =
  let m = branching_bit m p1 p2 in
  let p = mask p1 m
  in
    if p1 < p2
    then Branch (p, m, c, s1, s2)
    else Branch (p, m, c, s2, s1)
;;

let branch p m l r = Branch (p, m, size l + size r, l, r);;


let rec add s i x =
  if i < 0 then invalid_arg "IntMap.add: negative key";
  match s with
      Empty -> Leaf (i, x)
    | Leaf (j, _) when j = i -> raise Already_present
    | Leaf (j, _) -> join 1 i (Leaf (i, x)) j s 2
    | Branch (p, m, c, l, r) ->
	if prefix_matches i p m then
          if i <= p
          then branch p m (add l i x) r
          else branch p m l (add r i x)
	else
          join (m lsl 1) i (Leaf (i, x)) p s (c + 1)
;;

let rec set s i x =
  if i < 0 then invalid_arg "IntMap.set: negative key";
  match s with
      Empty -> Leaf (i, x)
    | Leaf (j, _) when j = i -> Leaf (i, x)
    | Leaf (j, _) -> join 1 i (Leaf (i, x)) j s 2
    | Branch (p, m, c, l, r) ->
	if prefix_matches i p m then
          if i <= p
          then branch p m (set l i x) r
          else branch p m l (set r i x)
	else
          join (m lsl 1) i (Leaf (i, x)) p s (c + 1)
;;

let rec delete s i =
  if i < 0 then invalid_arg "IntMap.delete: negative key";
  let branch = function
    | (_, _, Empty, s) -> s
    | (_, _, s, Empty) -> s
    | (p, m, l, r) -> Branch (p, m, size l + size r, l, r)
  in
    match s with
	Leaf (j, _) when j = i -> Empty
      | Branch (p, m, c, l, r) ->
          if prefix_matches i p m then
            if i <= p
            then branch (p, m, delete l i, r)
            else branch (p, m, l, delete r i)
          else
            s
      | _ -> raise Not_found
;;

let remove m i = try delete m i with Not_found -> m;;

let rec modify f s i =
  if i < 0 then invalid_arg "IntMap.modify: negative key";
  match s with
      Leaf (j, x) when j = i -> Leaf (i, f x)
    | Branch (p, m, _, l, r) ->
	if i <= p
	then branch p m (modify f l i) r
	else branch p m l (modify f r i)
    | _ -> raise Not_found
;;

let rec modify_default f s i x =
  if i < 0 then invalid_arg "IntMap.modify_default: negative key";
  match s with
      Empty -> Leaf (i, f x)
    | Leaf (j, y) when j = i -> Leaf (i, f y)
    | Leaf (j, _) -> join 1 i (Leaf (i, f x)) j s 2
    | Branch (p, m, c, l, r) ->
	if prefix_matches i p m then
          if i <= p
          then branch p m (modify_default f l i x) r
          else branch p m l (modify_default f r i x)
	else
          join (m lsl 1) i (Leaf (i, f x)) p s (c + 1)
;;

let rec fold_intersection f s1 s2 accu = match s1, s2 with
    Empty, _
  | _, Empty -> accu
  | Leaf (i, x), _ -> (match find s2 i with None -> accu | Some y -> f accu i x y)
  | _, Leaf (i, y) -> (match find s1 i with None -> accu | Some x -> f accu i x y)
  | Branch (p1, m1, _, l1, r1), Branch (p2, m2, _, l2, r2) ->
      if m1 = m2 && p1 = p2 then
	let accu = fold_intersection f l1 l2 accu in
	let accu = fold_intersection f r1 r2 accu in
	accu
      else if m1 > m2 && prefix_matches p2 p1 m1 then
        fold_intersection f (if p2 <= p1 then l1 else r1) s2 accu
      else if m1 < m2 && prefix_matches p1 p2 m2 then
        fold_intersection f s1 (if p1 <= p2 then l2 else r2) accu
      else
        accu
;;

let rec intersection s1 s2 =
  let disjoint_union s1 s2 = match s1, s2 with
    | _, Empty -> s1
    | Empty, _ -> s2
    | Leaf (i, x), _ -> add s2 i x
    | _, Leaf (i, x) -> add s1 i x
    | Branch (p1, m1, c1, l1, r1), Branch (p2, m2, c2, l2, r2) ->
      assert (not (prefix_matches p1 p2 m2));
      join (m1 lsl 1) p1 s1 p2 s2 (c1 + c2)
  in
  match s1, s2 with
  | Empty, _
  | _, Empty -> Empty
  | Leaf (i, x), _ -> if has_key s2 i then s1 else Empty
  | _, Leaf (i, x) -> (match find s1 i with None -> Empty | Some y -> Leaf (i, y))
  | Branch (p1, m1, _, l1, r1), Branch (p2, m2, _, l2, r2) ->
    if m1 = m2 && p1 = p2 then
      disjoint_union (intersection l1 l2) (intersection r1 r2)
    else if m1 > m2 && prefix_matches p2 p1 m1 then
      intersection (if p2 <= p1 then l1 else r1) s2
    else if m1 < m2 && prefix_matches p1 p2 m2 then
      intersection s1 (if p1 <= p2 then l2 else r2)
    else
      Empty
;;

let rec union_with f s1 s2 =
  let flip f x y = f y x in
  let rec update f s i x =
    match s with
    | Empty -> failwith "IntMap.union_with: internal error"
    | Leaf (j, y) when i = j -> Leaf (i, f x y)
    | Leaf (j, y) -> join 1 i (Leaf (i, x)) j s 2
    | Branch (p, m, c, l, r) ->
      if prefix_matches i p m then
	if i <= p
	then Branch (p, m, c + 1, (update f l i x), r)
	else Branch (p, m, c + 1, l, (update f r i x))
      else
	join (m lsl 1) i (Leaf (i, x)) p s (c + 1)
  in
  match s1, s2 with
  | Empty, t
  | t, Empty  -> t
  | Leaf (i, x), t -> update (flip f) t i x
  | t, Leaf (i, x) -> update f t i x
  | Branch (p1, m1, c1, l1, r1), Branch (p2, m2, c2, l2, r2) ->
      if m1 = m2 && p1 = p2 then
	(* The trees have the same prefix. Merge the subtrees.  *)
	branch p1 m1 (union_with f l1 l2) (union_with f r1 r2)
      else if m1 > m2 && prefix_matches p2 p1 m1 then
	(* [p2] contains [p1]. Merge [s2] with a subtree of [s1].  *)
	if p2 <= p1
	then branch p1 m1 (union_with f l1 s2) r1
	else branch p1 m1 l1 (union_with f r1 s2)
      else if m1 < m2 && prefix_matches p1 p2 m2 then
	(* [p1] contains [p2]. Merge [s1] with a subtree of [s2].  *)
	if p1 <= p2
	then branch p2 m2 (union_with f s1 l2) r2
	else branch p2 m2 l2 (union_with f s1 r2)
      else
	(* The prefixes disagree.  *)
	join (m1 lsl 1) p1 s1 p2 s2 (c1 + c2)
;;


(* TODO: optimize  *)
let filter p m =
  fold (fun m' k x -> if p k x then add m' k x else m') m empty
;;

let rec map f = function
    Empty -> Empty
  | Leaf (i, x) -> Leaf (i, f i x)
  | Branch (p, m, c, l, r) -> Branch (p, m, c, map f l, map f r)
;;

let output p channel m =
  Printf.fprintf channel "{[%d] " (size m);
  ignore (fold
    (fun first i x ->
       if not first then Printf.fprintf channel ", ";
       Printf.fprintf channel "%d:%a" i p x;
       false)
    m
    true);
  Printf.fprintf channel "}";
;;
