(* hcd -- solve the Highly Connected Deletion problem
   Copyright (C) 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

let usage_msg = "Solve the Hɪɢʜʟʏ Cᴏɴɴᴇᴄᴛᴇᴅ Dᴇʟᴇᴛɪᴏɴ problem";;

type solver =
  | CPLEX
  | GUROBI
  | COLGEN
;;

let stats_only        = ref false;;
let heuristic	      = ref false;;
let ilp_solver	      = ref GUROBI;;
let do_reduce	      = ref true;;
let comp_dump_file    = ref None;;
let cluster_file      = ref None;;
let time_limit	      = ref None;;

let specs = [
  ("-d",   Arg.String(fun f -> comp_dump_file := Some f),
           "Save largest connected component after reduction to file");
  ("-h",   Arg.Set(heuristic),
           "Heuristic solution");
  ("-i",   Arg.String(function
             | "cplex"  -> ilp_solver := CPLEX
             | "gurobi" -> ilp_solver := GUROBI
             | "colgen" -> ilp_solver := COLGEN
             | _ -> failwith "unknown ILP solver"),
           "ILP solver (cplex or gurobi)");
  ("-n",   Arg.Clear(do_reduce),
           "No data reduction");
  ("-o",   Arg.String(fun f -> cluster_file := Some f),
           "Save clusters to file instead of stdout");
  ("-s",   Arg.Set(stats_only),
           "Print statistics (suppresses printing clusters except when -o given)");
  ("-t",   Arg.Int(fun t -> time_limit := Some t),
           "Print statistics (suppresses printing clusters except when -o given)");
  ("-v",   Arg.Set(Util.verbose),
           "Verbose output");
];;

external min_cut_c : (int * int) array -> (int * int array) = "c_min_cut";;

external hcd_ilp_c : int -> (int * int) array -> solver -> int -> bool -> (int * int) array = "c_hcd_ilp";;

let min_cut g =
  if Graph.num_vertices g < 2 then IntSet.empty, max_int else
  let old_of_new, new_of_old, n = Graph.fold_vertices
    (fun (old_of_new, new_of_old, i) u _ ->
      (IntMap.add old_of_new i u),
      (IntMap.add new_of_old u i),
      (i + 1))
    g (IntMap.empty, IntMap.empty, 0) in
  let edges = Graph.fold_edges
    (fun edges u v -> (IntMap.get new_of_old u, IntMap.get new_of_old v) :: edges) g [] in
  let edges = Array.of_list edges in
  let w, cut = min_cut_c edges in
  let cut = Array.map (fun u -> IntMap.get old_of_new u) cut in
  Array.fold_left (fun cut u -> IntSet.add cut u) IntSet.empty cut, w
;;

let make_g_x g =
  Graph.fold_vertex_pairs
    (fun g_x u n_u v n_v ->
      if Graph.is_connected g u v || IntSet.intersection_size n_u n_v >= 3
      then Graph.connect g_x u v
      else g_x)
    g (Graph.fold_vertices (fun g_x u _ -> Graph.add_vertex g_x u) g Graph.empty)
;;

let fold_k3s f g z =
  Graph.fold_edges
    (fun z u v ->
      let n_uv = IntSet.intersection (Graph.neighbors g u) (Graph.neighbors g v) in
      IntSet.fold (fun z w -> if w > v then f z u v w else z) n_uv z)
    g z
;;

let reduce_k3s g =
  let smalldeg = Graph.fold_vertices
    (fun vs u n_u -> if IntSet.size n_u <= 10 then IntSet.add vs u else vs)
    g IntSet.empty in
  fold_k3s
    (fun (g, dk) u v w ->
      if not (Graph.is_connected g u v) || not (Graph.is_connected g u w) || not (Graph.is_connected g v w)
      then g, dk
      else
	let n_u = Graph.neighbors g u in
	let n_v = Graph.neighbors g v in
	let n_w = Graph.neighbors g w in
	let n_u = List.fold_left IntSet.delete n_u [v; w] in
	let n_v = List.fold_left IntSet.delete n_v [u; w] in
	let n_w = List.fold_left IntSet.delete n_w [u; v] in
	let largest_cc vs =
	  Graph.fold_connected_components
	    (fun size cc -> max size (Graph.num_vertices cc))
	    (Graph.subgraph g vs) 0 in
	let edges_between vs1 vs2 =
	  IntSet.fold (fun yes x -> yes || IntSet.do_intersect (Graph.neighbors g x) vs2) vs1 false in
	let u_gain = largest_cc n_u in
        let v_gain = largest_cc n_v in
	let w_gain = largest_cc n_w in
	let u_gain = if u_gain = 1 then 0 else u_gain in
	let v_gain = if v_gain = 1 then 0 else v_gain in
	let w_gain = if w_gain = 1 then 0 else w_gain in
	let n_uv = IntSet.intersection n_u n_v in
	let n_uw = IntSet.intersection n_u n_w in
	let n_vw = IntSet.intersection n_v n_w in
	let uv_gain = 2 * largest_cc n_uv in
	let uw_gain = 2 * largest_cc n_uw in
	let vw_gain = 2 * largest_cc n_vw in
	if u_gain + v_gain + w_gain <= 3
	  && u_gain + vw_gain <= 3
	  && v_gain + uw_gain <= 3
	  && w_gain + uv_gain <= 3
	  && IntSet.is_empty (IntSet.intersection n_u n_vw)
	  && not (edges_between n_uv n_uw)
	  && not (edges_between n_uv n_vw)
	  && not (edges_between n_uw n_vw)
	then 
	  let del = IntSet.fold (fun del x -> (u, x) :: del) n_u [] in
	  let del = IntSet.fold (fun del x -> (v, x) :: del) n_v del in
	  let del = IntSet.fold (fun del x -> (w, x) :: del) n_w del in
	  List.fold_left (fun (g, dk) (x, y) -> Graph.disconnect g x y, dk + 1) (g, dk) del
	else
	  g, dk)
    (Graph.subgraph g smalldeg) (g, 0)
;;

let reduce_triangles g =
  Graph.fold_vertices
    (fun (g, dk) u n_u ->
      let n_u_deg2 = IntSet.filter (fun v -> Graph.deg g v = 2) n_u in
      let g' = Graph.subgraph g n_u_deg2 in
      if Graph.num_edges g' > 0 then
	let isol = Graph.fold_vertices
	  (fun isol v n_v -> if IntSet.size n_v = 0 then IntSet.add isol v else isol)
	  g' IntSet.empty in
	if not (IntSet.is_empty isol)
	then
	  IntSet.fold
	    (fun (g, dk) v -> Graph.fold_neighbors (fun g w -> Graph.disconnect g v w) g v g, dk + 2)
	    isol (g, dk)
	else
	  g, dk
      else begin
	let n_n_u_deg2 = IntSet.fold
	  (fun n_n_u_deg2 u -> IntSet.union n_n_u_deg2 (Graph.neighbors g u))
	  n_u_deg2 IntSet.empty in
	let n_n_u_deg2 = IntSet.minus n_n_u_deg2 n_u_deg2 in
	let ns = IntSet.intersection n_u n_n_u_deg2 in 
	let ns3 = IntSet.filter (fun v -> Graph.deg g v = 3) ns in
	if not (IntSet.is_empty ns3) then begin
	  Printf.eprintf "u = %d ns3 = %a\n%!" u IntSet.output ns3;
	  (* Graph.output (Graph.subgraph g (IntSet.add n_u u)); *)
	  IntSet.fold
	    (fun (g, dk) v ->
	      let n_v = Graph.neighbors g v in
	      let n_v = IntSet.delete n_v u in
	      assert (IntSet.size n_v = 2);
	      let n_v = IntSet.intersection n_v n_u_deg2 in
	      assert (IntSet.size n_v = 1);
	      let w = IntSet.choose n_v in
	      Graph.disconnect g v w, dk + 1)
	    ns3 (g, dk)
	end else
	  g, dk
      end
	)
    g (g, 0)
;;

let rec reduce g =
  if !Util.verbose then Printf.eprintf "reducing, n = %d m = %d\n%!" (Graph.num_vertices g) (Graph.num_edges g);
  if not !do_reduce then (g, 0) else
    let g, dk =
      Graph.fold_edges
	(fun (g, dk) u v ->
	  let n_u = Graph.neighbors g u in
	  let n_v = Graph.neighbors g v in
	  if IntSet.do_intersect n_u n_v then g, dk
	  else Graph.disconnect g u v, dk + 1)
	g (g, 0) in
    let gx = make_g_x g in
    let del =
      Graph.fold_vertex_pairs
	(fun del u n_u v n_v ->
	  if not (IntSet.is_empty n_u) && IntSet.size n_u <= 3 && n_u = n_v && not (Graph.is_connected gx u v)
	  then IntSet.put del v
	  else del)
	g IntSet.empty in
    let g, dk =
      IntSet.fold
	(fun (g, dk) u -> Graph.add_vertex (Graph.delete_vertex g u) u, dk + Graph.deg g u)
	del (g, dk) in
    let g, dk' = reduce_k3s g in
    if dk' > 0 then let g, dk'' = reduce g in g, dk + dk' + dk''
    else
      let g, dk' = reduce_triangles g in
      if dk' > 0 then let g, dk'' = reduce g in g, dk + dk' + dk''
      else
	g, dk
;;

let wgraph_of_graph g =
  let wg = Graph.fold_vertices (fun wg u _ ->  WGraph.add_vertex wg u) g WGraph.empty in
  Graph.fold_edges (fun wg u v -> WGraph.connect wg u v 1) g wg
;;

let rec hcs_heuristic f g z =
  if !Util.verbose && Graph.num_vertices g >= 3 then Printf.eprintf "heuristic n = %5d m = %6d%!" (Graph.num_vertices g) (Graph.num_edges g);
  let vs = Graph.vertex_set g in
  (* let cut, weight = WGraph.min_cut (wgraph_of_graph g) in *)
  let cut, weight = min_cut g in
  if !Util.verbose && Graph.num_vertices g >= 3 then Printf.eprintf " cut %d\n%!" weight;
  if weight > Graph.num_vertices g / 2
  then f z vs
  else
    let g1, _ = reduce (Graph.subgraph g cut) in
    let g2, _ = reduce (Graph.subgraph g (IntSet.minus vs cut)) in
    let z = Graph.fold_connected_components (fun z g' -> hcs_heuristic f g' z) g1 z in
    let z = Graph.fold_connected_components (fun z g' -> hcs_heuristic f g' z) g2 z in
    z
;;

let hcd_ilp f g z =
  if Graph.num_vertices g = 1 then f z (Graph.vertex_set g) else
  let old_of_new, new_of_old, n = Graph.fold_vertices
    (fun (old_of_new, new_of_old, i) u _ ->
      (IntMap.add old_of_new i u),
      (IntMap.add new_of_old u i),
      (i + 1))
    g (IntMap.empty, IntMap.empty, 0) in
  let edges = Graph.fold_edges
    (fun edges u v -> (IntMap.get new_of_old u, IntMap.get new_of_old v) :: edges) g [] in
  let edges = Array.of_list edges in
  let g' = Graph.fold_vertices (fun g' u _ -> Graph.add_vertex g' (IntMap.get new_of_old u)) g Graph.empty in
  let g' = Graph.fold_edges (fun g' u v -> Graph.connect g' (IntMap.get new_of_old u) (IntMap.get new_of_old v)) g g' in
  Heuristic.current_g := g';
  let del = hcd_ilp_c (Graph.num_vertices g) edges !ilp_solver
    (match !time_limit with None -> 0 | Some t -> t) !Util.verbose in
  let del = Array.map (fun (u, v) -> IntMap.get old_of_new u, IntMap.get old_of_new v) del in
  let g' = Array.fold_left (fun g' (u, v) -> Graph.disconnect g' u v) g del in
  Graph.fold_connected_components (fun z g -> f z (Graph.vertex_set g)) g' z
;;

let is_hcc g cc =
  let g' = Graph.subgraph g cc in
  let _, w = min_cut g' in
  (float_of_int w) > (0.5 *. (float_of_int (Graph.num_vertices g')))
;;

let verify g hccs =
  let components =
    IntMap.fold
      (fun components u c_u ->
	IntMap.modify_default (fun s -> IntSet.add s u) components c_u IntSet.empty)
      hccs IntMap.empty in
  IntMap.iter
    (fun _ cc ->
      if not (is_hcc g cc)
      then begin
	Printf.eprintf "internal error: %a is not highly connected\n%!" IntSet.output cc;
	exit 1;
      end)
    components
;;

let cc_num   = ref 0;;		(* number of connected components in input *)
let cc_max_n = ref 0;;		(* largest connected component in input *)
let cc_max_m = ref 0;;

let dr_dk    = ref 0;;		(* number of edges removed in DR *)
let dr_num   = ref 0;;		(* number of connected components after DR *)
let dr_max_n = ref 0;;		(* largest connected component after DR *)
let dr_max_m = ref 0;;

let hcc_triv    = ref 0;;	(* number of trivial hc components *)
let hcc_nontriv = ref 0;;	(* number of nontrivial hc components *)
let hcc_max_n	= ref 0;;	(* largest hc component *)
let hcc_max_m	= ref 0;;

let main () =
  Arg.parse specs (fun _ -> Arg.usage specs usage_msg) usage_msg;
  let g, vertex_names, vertex_numbers = Graph.input stdin in
  let clusters_out = match !cluster_file with None -> stdout | Some f -> open_out f in
  let handle_hcc (hccs, c) hcc =
    if IntSet.size hcc > 1 && (not !stats_only || !cluster_file <> None) then begin
      ignore (IntSet.fold (fun first i ->
	Printf.fprintf clusters_out "%s%s"
	  (if first then "" else " ")
	  (IntMap.get vertex_names i);
	false) hcc true);
      Printf.fprintf clusters_out "\n";
    end;
    
    if IntSet.size hcc = 1 then incr hcc_triv  else incr hcc_nontriv;
    if IntSet.size hcc > !hcc_max_n then begin
	hcc_max_n := IntSet.size hcc;
	hcc_max_m := Graph.num_edges (Graph.subgraph g hcc);
    end;
    
    let hccs = IntSet.fold (fun hccs u -> IntMap.add hccs u c) hcc hccs in
    hccs, c + 1
  in
  let start = Util.timer () in
  let (hccs, _) =
    Graph.fold_connected_components
      (fun z cc ->
	if !Util.verbose && Graph.num_vertices cc >= 3 then Printf.eprintf "component n = %5d m = %6d%!" (Graph.num_vertices cc) (Graph.num_edges cc);
	incr cc_num;
	if Graph.num_vertices cc > !cc_max_n then begin
	  cc_max_n := Graph.num_vertices cc;
	  cc_max_m := Graph.num_edges cc;
	end;
	let cc, dk = reduce cc in
	dr_dk := !dr_dk + dk;
	if !Util.verbose && Graph.num_vertices cc >= 3 then Printf.eprintf " reduced %6d edges\n%!" dk;
	Graph.fold_connected_components
	  (fun z cc ->
	    if !Util.verbose && Graph.num_vertices cc >= 3 then Printf.eprintf "subcomp.  n = %5d m = %6d\n%!" (Graph.num_vertices cc) (Graph.num_edges cc);
	    incr dr_num;
	    if Graph.num_vertices cc > !dr_max_n then begin
	      dr_max_n := Graph.num_vertices cc;
	      dr_max_m := Graph.num_edges cc;
	      match !comp_dump_file with
	      | None -> ()
	      | Some f ->
		if !Util.verbose then Printf.eprintf "saving reduced component to '%s'\n%!" f;
		let o = open_out f in
		Graph.output o cc;
		close_out o;
	    end;
	    (if !heuristic then hcs_heuristic else hcd_ilp) handle_hcc cc z)
	  cc z)
      g (IntMap.empty, 0) in
  let stop = Util.timer () in
  if !cluster_file <> None then close_out clusters_out;
  verify g hccs;
  let n = Graph.num_vertices g in
  let m = Graph.num_edges g in
  let k = Graph.fold_edges (fun k u v -> if IntMap.get hccs u <> IntMap.get hccs v then k + 1 else k) g 0 in
  if !stats_only then
    Printf.printf
      "%5d %6d %6d  %4d %5d %6d  %6d %4d %5d %6d  %5d %4d %5d %6d %8.2f\n%!"
      n m k
      !cc_num !cc_max_n !cc_max_m
      !dr_dk !dr_num !dr_max_n !dr_max_m
      !hcc_triv !hcc_nontriv !hcc_max_n !hcc_max_m
      (stop -. start)
;;

(* Subgraph of g containing vertices and edges that might be in a
   highly connected component that contains u.  *)
let rec hcd_g_u g u =
  let n_u = Graph.neighbors g u in
  let n2_u = IntSet.fold (fun n2_u v -> IntSet.union n2_u (Graph.neighbors g v)) n_u IntSet.empty in
  let n2_u = IntSet.minus n2_u n_u in
  let n2_u = IntSet.remove n2_u u in
  let n2_u = IntSet.filter (fun v -> IntSet.intersection_size n_u (Graph.neighbors g v) >= 3) n2_u in
  let g' = Graph.subgraph g (IntSet.put (IntSet.union n_u n2_u) u) in
  let g'', _ =
    Graph.fold_edges
      (fun (g, dk) u v ->
	let n_u = Graph.neighbors g u in
	let n_v = Graph.neighbors g v in
	if IntSet.do_intersect n_u n_v then g, dk
	else Graph.disconnect g u v, dk + 1)
      g' (g', 0) in
  (* Printf.eprintf "%7d %4d %4d\n%!" u (Graph.num_vertices g') (Graph.num_edges g'); *)
  if g'' <> g then hcd_g_u g'' u
  else g''
;;

(*

let main() =
  let g, vertex_names, vertex_numbers = Graph.input stdin in
  Printf.printf "n = %d m = %d\n%!" (Graph.num_vertices g) (Graph.num_edges g);
  let rec loop g =
      if Graph.num_vertices g = 0 then () else
	(*
	let u, n2_u = Graph.fold_vertices
	  (fun (u_min, n2_u_min) u n_u ->
	    let n2_u = IntSet.fold (fun n2_u v -> IntSet.union n2_u (Graph.neighbors g v)) n_u n_u in
	    let n2_u = IntSet.put n2_u u in
	    if u_min = -1 || IntSet.size n2_u < IntSet.size n2_u_min
	    then u, n2_u
	    else u_min, n2_u_min)
	  g (-1, IntSet.empty) in
	let g' = Graph.subgraph g n2_u in
	let g'', dk =
	  Graph.fold_edges
	    (fun (g, dk) u v ->
	      let n_u = Graph.neighbors g u in
	      let n_v = Graph.neighbors g v in
	      if IntSet.do_intersect n_u n_v then g, dk
	      else Graph.disconnect g u v, dk + 1)
	    g' (g', 0) in
	Printf.printf "u = %7d n' = %4d m' = %4d n'' = %4d m'' = %4d\n%!"
	  u
	  (Graph.num_vertices g') (Graph.num_edges g')
	  (Graph.num_vertices g'') (Graph.num_edges g'');
	*)
	let u, g_u = Graph.fold_vertices
	  (fun (min_u, min_g_u) u _ ->
	    let g_u = hcd_g_u g u in
	    if min_u = -1 || Graph.num_vertices g_u < Graph.num_vertices min_g_u
	    then u, g_u
	    else min_u, min_g_u)
	  g (-1, Graph.empty) in
	Printf.printf "u = %7d n' = %4d m' = %4d\n%!" u (Graph.num_vertices g_u) (Graph.num_edges g_u);
	if Graph.num_vertices g_u > 1 then begin
	  let f = open_out (Printf.sprintf "g_%07d.edges" u) in
	  Graph.output f g_u;
	  close_out f;
	end;
	loop (Graph.delete_vertex g u)
  in
  loop g
;;

*)
  
main ();;
