(* hcd -- solve the Highly Connected Deletion problem
   Copyright (C) 2011, 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

(** Max priority search queues, implemented as big-endian patricia tree
    with "winner" annotations.  *)

type priority = float;;

type t =
    Empty
  | Leaf of int * priority
  | Branch of int * int * priority * t * t
    (** [Branch (p, m, c, l, r)]:
        [p] is the largest common prefix for all the keys in this tree
        [m] is the branching bit mask
            ([m] is a power of 2, only the bits above [m] are valid in [p])
        [c] is the maximum priority in this subtree
        [l] contains all the keys with a 0 in the branching bit
        [r] contains all the keys with a 1 in the branching bit  *)
;;

let empty = Empty;;
let is_empty s = s = Empty;;

(* Return an integer where only the highest bit that was set in [x] is
   still set.  *)
let branching_bit m p1 p2 =
  let rec highest_bit x =
    let x' = x land (x - 1) in
      if x' = 0
      then x
      else highest_bit x' in
  let x = p1 lxor p2 in			(* remove common prefix *)
  let x = x land lnot (m - 1) in	(* remove invalid suffix *)
    highest_bit x
;;

(* In [i], clear the 1-bit mask [m], and set all bits below [m]'s bit to one.  *)
let mask i m = (i lor (m - 1 + m)) - m;;

let prefix_matches i p m = (mask i m = p)

let priority = function
  | Empty -> failwith "PSQueue.priority"
  | Leaf (_, pri) -> pri
  | Branch (_, _, pri, _, _) -> pri
;;
  
let branch p m l r =
  Branch (p, m, max (priority l) (priority r), l, r)
;;
  
let join m p1 s1 p2 s2 =
  let m = branching_bit m p1 p2 in
  let p = mask p1 m
  in
    if p1 < p2
    then branch p m s1 s2
    else branch p m s2 s1
;;

let rec add s i pri =
  if i < 0 then invalid_arg "PSQueue.add: negative key";
  match s with
  | Empty -> Leaf (i, pri)
  | Leaf (j, _) when j = i -> failwith "PSQueue.add: element already contained"
  | Leaf (j, _) -> join 1 i (Leaf (i, pri)) j s
  | Branch (p, m, _, l, r) ->
      if prefix_matches i p m then
	if i <= p
	then branch p m (add l i pri) r
	else branch p m l (add r i pri)
      else
	join (m lsl 1) i (Leaf (i, pri)) p s
;;

let rec update f s i = match s with
  | Empty -> Leaf (i, f None)
  | Leaf (j, pri) when j = i -> Leaf (i, f (Some pri))
  | Leaf (j, _) -> join 1 i (Leaf (i, f None)) j s
  | Branch (p, m, _, l, r) ->
    if prefix_matches i p m then
      if i <= p
      then branch p m (update f l i) r
      else branch p m l (update f r i)
    else
      join (m lsl 1) i (Leaf (i, f None)) p s
;;

(* update only if present  *)
let rec update' f s i = match s with
  | Leaf (j, pri) when j = i -> Leaf (i, f pri)
  | Branch (p, m, _, l, r) when prefix_matches i p m ->
    if i <= p
    then branch p m (update' f l i) r
    else branch p m l (update' f r i)
  | _ -> s
;;

let pop s =
  let rec pop' = function
    | Empty -> failwith "PSQueue.pop: internal error"
    | Leaf (i, pri) -> i, pri, Empty
    | Branch (p, m, _, l, r) ->
      if priority l > priority r
      then let i, pri, l = pop' l in i, pri, if l = Empty then r else branch p m l r
      else let i, pri, r = pop' r in i, pri, if r = Empty then l else branch p m l r
  in
  if s = Empty then None else Some (pop' s)
;;
