(* hcd -- solve the Highly Connected Deletion problem
   Copyright (C) 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)

type vertex = int;;
type weight = int;;
type vertex_set = IntSet.t;;
type neighbors = weight IntMap.t;;
type t = neighbors IntMap.t;;

let empty = IntMap.empty;;


(* Querying  *)

let has_vertex g u = IntMap.has_key g u;;
let min_vertex g = IntMap.min_key g;;
let num_vertices g = IntMap.size g;;

let neighbors g u = 
  assert (has_vertex g u);
  IntMap.get g u
;;

let is_connected g u v = 
  assert (has_vertex g u);
  assert (has_vertex g v);
  IntMap.has_key (neighbors g u) v
;;

let fold_vertices = IntMap.fold;;
let iter_vertices = IntMap.iter;;

let fold_neighbors f g u z =
  assert (has_vertex g u);
  IntMap.fold f (neighbors g u) z
;;

let fold_edges f g z = 
  fold_vertices (fun z u n_u -> IntMap.fold (fun z v w -> if u < v then f z u v w else z) n_u z) g z
;;

let iter_edges f g = fold_edges (fun () u v w -> f u v w) g ();;

let num_edges g =
  let num = fold_vertices (fun num _ n_u -> num + IntMap.size n_u) g 0 in
  assert (num mod 2 = 0);
  num / 2
;;

let merge g u v =
  assert (has_vertex g u);
  assert (has_vertex g v);
  let n_v = neighbors g v in
  let g = fold_neighbors
    (fun g w w_vw ->
      IntMap.modify
	(fun n_w -> IntMap.modify_default ((+) w_vw) (IntMap.delete n_w v) u 0)
	g w)
    g v g in
  let g = IntMap.delete g v in
  IntMap.modify (fun n_u -> IntMap.remove (IntMap.union_with (+) n_u n_v) u) g u
;;

let fold_connected_components f g z =
  let rec dfs seen component u =
    fold_neighbors
      (fun (seen, component) v _ ->
	if IntSet.contains seen v
	then seen, component
	else
	  let seen = IntSet.add seen v in
	  let component = IntMap.add component v () in
	  dfs seen component v)
      g u (seen, component)
  in
  fst
    (fold_vertices
       (fun (z, seen) u _ ->
	 if IntSet.contains seen u
	 then (z, seen)
	 else
	   let seen, component = dfs (IntSet.add seen u) (IntMap.singleton u ()) u in
	   f (IntMap.intersection g component) z, seen)
       g (z, IntSet.empty))
;;

let rec min_cut g =
  (* Printf.fprintf stderr "min_cut n = %d\n%!" (num_vertices g); *)
  let find_st_cut g =
    let rec loop q s t w =
      match PSQueue.pop q with
      | None -> (s, t, IntSet.singleton t, int_of_float w)
      | Some (u, w_u, q) ->
	assert (has_vertex g u);
	loop (fold_neighbors (fun q v w_uv -> PSQueue.update' ((+.) (float_of_int w_uv)) q v) g u q) t u w_u
    in
    loop (fold_vertices (fun q u _ -> PSQueue.add q u 0.0) g PSQueue.empty) (-1) (-1) infinity
  in
  if num_vertices g < 2
  then IntSet.empty, max_int
  else
    let s, t, st_cut, st_cut_weight = find_st_cut g in
    let cut', cut_weight' = min_cut (merge g s t) in
    if st_cut_weight < cut_weight'	(* prefer non-size-1 cut *)
    then st_cut, st_cut_weight
    else (if IntSet.contains cut' s then IntSet.add cut' t else cut'), cut_weight'
;;


(*  Modifications  *)

let add_vertex g u =
  assert (not (has_vertex g u));
  IntMap.add g u IntMap.empty
;;


let connect g u v w =
  assert (has_vertex g u);
  assert (has_vertex g v);
  assert (not (is_connected g u v));
  let g = IntMap.modify (fun n_u -> IntMap.add n_u v w) g u in
  let g = IntMap.modify (fun n_v -> IntMap.add n_v u w) g v in
  g
;;


(* I/O  *)

let output channel g =
  (* IntMap.output (IntMap.output Util.output_int) channel g *)
  Printf.fprintf channel "{ n = %d, m = %d\n" (num_vertices g) (num_edges g);
  iter_vertices (fun i neighbors ->
                   if IntMap.is_empty neighbors
                   then Printf.fprintf channel "%d\n" i) g;
  iter_edges (fun i j w -> Printf.fprintf channel "%d %d %d\n" i j w) g;
  Printf.fprintf channel "}\n";
;;
