/* hcd -- solve the Highly Connected Deletion problem
   Copyright (C) 2012, 2013  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  */

#if 0

#include <cmath>
#include <set>
#include <tuple>
#include <vector>

#include "CGraph.hh"
#include "UGraph.hh"
#include "lp/LP.hh"
#include "lp/LPCplex.hh"
#include "Timer.hh"

std::pair<double, std::set<vertex_t>>
colgenHeuristic(const Graph& g, const std::vector<double>& vertexWeights);

std::pair<double, std::set<vertex_t>>
independentSetHeuristic(const std::vector<double>& vertexWeights);

LP::Solver solver = LP::Solver::CPLEX;

Timer timerILP, timerCuttingPlanes;

class Callback : public LP::Callback {
public:
    Callback(const std::vector<LP::Var>& nauxVertexVars) : auxVertexVars(nauxVertexVars) { }
    void call(LP::CallbackModel& model) {
	timerILP.stop();
	timerCuttingPlanes.start();
	std::size_t n = auxVertexVars.size();
	std::set<vertex_t> is;
	double isValue;
	std::vector<double> weights(n);
	for (size_t i = 0; i < n; ++i)
	    weights[i] = model.value(auxVertexVars[i]);
	std::tie(isValue, is) = independentSetHeuristic(weights);
	if (isValue > 1.05) {
	    //if (isValue >= 1.1) {
	    //std::cerr << "found IS of size " << is.size() << " with weight " << isValue << std::endl;
	    LP::Expr sum;
	    for (vertex_t u : is)
		sum += auxVertexVars[u];
	    //std::cerr << (sum <= 1) << std::endl;
	    model.addUserCut(sum <= 1);
	}
	timerCuttingPlanes.stop();
	timerILP.start();
    }

private:
    const std::vector<LP::Var>& auxVertexVars;
};

std::vector<UGraph> calcN2s(const Graph& g0) {
    UGraph g;
    std::vector<UGraph> gs(g0.num_vertices());
    for (vertex_t u = 0; u < g0.num_vertices(); u++)
	g.addVertex(u);
    for (vertex_t u = 0; u < g0.num_vertices(); u++) {
	for (NeighborIt ni = g0.neighbors_begin(u); ni != g0.neighbors_end(u); ++ni) {
	    vertex_t v = ni.n();
	    if (u < v)
		g.addEdge(u, v);
	}
    }

    for (auto& u : g) {
	UGraph::Neighbors n2{u.second};
	n2.insert(u.first);
	for (UGraph::Vertex v : u.second) {
	    const UGraph::Neighbors& n_v = g.neighbors(v);
	    n2.insert(n_v.lower_bound(u.first + 1), n_v.end());
	}
	gs[u.first] = g.subgraph(n2);
	fprintf(stderr, "gs[%4zd]: n = %4zd m = %5zd\n",
		u.first, gs[u.first].numVertices(), gs[u.first].numEdges());
    }
    return gs;
}

std::pair<std::vector<UGraph::Vertex>, double> maxWeightHC(const UGraph& g, std::vector<double> w) {
    LP::Model* pauxModel = LP::newModel(solver);
    LP::Model& auxModel(*pauxModel);
    auxModel.setObjectiveSense(LP::Model::Sense::MAXIMIZE);
    std::map<UGraph::Vertex, LP::Var> auxVertexVars;
    // y_v = 1 <=> vertex v belongs to the cluster maximizing the objective
    for (auto& u : g) {
	sprintf(buf, "y_%zd", u.first);
	auxVertexVars[u.first] = auxModel.addBoolVar(buf, weight[u.first]);
    }
    // cluster size constraint
    LP::Var d = auxModel.addIntVar(0, LP::inf, "d");
    LP::Expr sum;
    for (size_t i = 0; i < n; i++)
	sum += auxVertexVars[i];

    // x_e: edge e belongs to the cluster maximizing the objective
    for (vertex_t u = 0; u < g.num_vertices(); u++) {
	for (NeighborIt ni = g.neighbors_begin(u); ni != g.neighbors_end(u); ++ni) {
	    vertex_t v = ni.n();
	    if (!(u < v))
		continue;
	    sprintf(buf, "x_%d_%d", u, v);
	    LP::Var xuv = auxModel.addBoolVar(-1, buf);
	    auxModel.addConstraint(xuv <= auxVertexVars[u]);
	    auxModel.addConstraint(xuv <= auxVertexVars[v]);
	}
    }

    // constraints that enforce highly connectedness
    for (size_t u = 0; u < n; u++) {
	LP::Expr sum;
	for (NeighborIt ni = g.neighbors_begin(u); ni != g.neighbors_end(u); ++ni)
	    sum += auxVertexVars[ni.n()];
	double bigM = 0.5 * n + 0.5;
	//auxModel.addConstraint(sum - 0.5 * d + bigM * (1-auxVertexVars[u]) >= 0.5);
	auxModel.addConditionalConstraint(true, auxVertexVars[u], bigM, sum >= 0.5 * d + 0.5);
    }
    auxModel.setGoodEnough(-1e-6);
    auxModel.setEmphasis(LP::Emphasis::FIND_FEASIBLE);

    auxModel.solve();
    if (auxModel.status() != LP::Status::OPTIMAL && auxModel.status() != LP::Status::ABORTED_FEASIBLE) {
	fprintf(stderr, "bad aux staus\n");
	exit(1);
    }

    std::set<vertex_t> cluster;
    for (auto& u : g) {
	double x = auxVertexVars[u.first].value();
	if (!((fabs(x - 0.0) < EPS || fabs(x - 1.0) < EPS))) {
	    fprintf(stderr, "x[%zd] = %f\n", u.first, x);
	    exit(1);
	}
	if (x > 0.5)
	    cluster.insert(u.first);
    }

    delete pauxModel;

    return std::make_pair(cluster, auxModel.objectiveValue());
}

std::vector<Edge> hcd_colgen(const Graph& g, int timeLimit, bool verbose) {
    Timer timer;
    timer.start();
    //std::vector<UGraph> gs = calcN2s(g);
    size_t n = g.num_vertices();

    // main problem
    LP::Model* pmodel = LP::newModel(solver);
    LP::Model& model(*pmodel);
    std::vector<LP::Var> clusterVars;
    std::vector<std::set<vertex_t> > clusters;
    std::vector<LP::Constraint> vertexConstraints;

    // initial z variables and constraints
    char buf[64];
    for (size_t i = 0; i < n; i++) {
	sprintf(buf, "z_%zd", i);
	LP::Var zi = model.addVar(0.0, 1.0, -0.5, buf);
	clusterVars.push_back(zi);
	LP::Constraint r = model.addConstraint(zi == 1);
	vertexConstraints.push_back(r);
	std::set<vertex_t> cluster;
	cluster.insert(i);
	clusters.push_back(cluster);
    }

    // auxiliary MIP
    LP::Model* pauxModel = LP::newModel(solver);;
    LP::Model& auxModel = *pauxModel;
    std::vector<LP::Var> auxVertexVars;
    // y_v = 1 <=> vertex v belongs to the cluster maximizing the objective
    for (size_t i = 0; i < n; i++) {
	sprintf(buf, "y_%zd", i);
	LP::Var yi = auxModel.addBoolVar(buf);
	auxVertexVars.push_back(yi);
    }
    // cluster size constraint
    LP::Var d = auxModel.addIntVar(0, LP::inf, "d");
    LP::Expr sum;
    for (size_t i = 0; i < n; i++)
	sum += auxVertexVars[i];
    auxModel.addConstraint(sum == d);
    //auxModel.addConstraint(d >= 3); // cutting plane, not needed for correctness

    // x_e: edge e belongs to the cluster maximizing the objective
    for (vertex_t u = 0; u < g.num_vertices(); u++) {
	for (NeighborIt ni = g.neighbors_begin(u); ni != g.neighbors_end(u); ++ni) {
	    vertex_t v = ni.n();
	    if (!(u < v))
		continue;
	    sprintf(buf, "x_%d_%d", u, v);
	    LP::Var xuv = auxModel.addBoolVar(-1, buf);
	    auxModel.addConstraint(xuv <= auxVertexVars[u]);
	    auxModel.addConstraint(xuv <= auxVertexVars[v]);
	}
    }

    // constraints that enforce highly connectedness
    for (size_t u = 0; u < n; u++) {
	LP::Expr sum;
	for (NeighborIt ni = g.neighbors_begin(u); ni != g.neighbors_end(u); ++ni)
	    sum += auxVertexVars[ni.n()];
	double bigM = 0.5 * n + 0.5;
	//auxModel.addConstraint(sum - 0.5 * d + bigM * (1-auxVertexVars[u]) >= 0.5);
	auxModel.addConditionalConstraint(true, auxVertexVars[u], bigM, sum >= 0.5 * d + 0.5);
    }

    auxModel.setGoodEnough(-1e-6);
    auxModel.setEmphasis(LP::Emphasis::FIND_FEASIBLE);
    //Callback callback(auxVertexVars);
    //auxModel.setCutCallback(callback, true, false);

    // generate columns
    for (;;) {
	if (timeLimit && timer.current() > timeLimit) {
	    fprintf(stderr, "spent %f seconds, bailing out...", timer.current());
	    break;
	}
	if (verbose)
	    fprintf(stderr, "%6zd columns; ", model.numVariables());
	model.solve();

	std::vector<double> vertexWeights(n);
	for (size_t i = 0; i < n; i++)
	    vertexWeights[i] = -vertexConstraints[i].dualValue();
	std::set<vertex_t> cluster;
	double clusterValue;
	std::tie(clusterValue, cluster) = colgenHeuristic(g, vertexWeights);

	constexpr double EPS = 1e-6;
	bool usedILP = false;
	if (clusterValue >= -EPS) {
	    usedILP = true;
	    if (verbose)
		fprintf(stderr, "ILP       ");
	    for (size_t i = 0; i < n; i++)
		auxModel.setObjectiveCoeff(auxVertexVars[i], -vertexConstraints[i].dualValue());
	    timerILP.reset();
	    timerCuttingPlanes.reset();
	    timerILP.start();
#if 0
	    timerILP.stop();
	    std::vector<vertex_t> weights(n);
	    for (size_t i = 0; i < n; i++)
		weights[i] = -vertexConstraints[i].dualValue();
	    std::tie(cluster, clusterValue) = maxWeightHC(g, weights);
#else
	    auxModel.solve();
	    timerILP.stop();
	    if (auxModel.status() != LP::Status::OPTIMAL && auxModel.status() != LP::Status::ABORTED_FEASIBLE) {
		fprintf(stderr, "bad aux staus\n");
		exit(1);
	    }
	    clusterValue = auxModel.objectiveValue();
	    cluster.clear();
	    for (size_t u = 0; u < n; u++) {
		double x = auxVertexVars[u].value();
		if (!((fabs(x - 0.0) < EPS || fabs(x - 1.0) < EPS))) {
		    fprintf(stderr, "x[%zd] = %f\n", u, x);
		    exit(1);
		}
		if (x > 0.5)
		    cluster.insert(u);
	    }
#endif
	    if (verbose)
		fprintf(stderr, "%9.4f", fabs(clusterValue));
	} else {
	    if (verbose)
		fprintf(stderr, "heuristic %9.4f", fabs(clusterValue));
	}	    
	if (clusterValue >= -EPS)
	    break;

	// value of cluster
	int value = 0;
	for (std::set<vertex_t>::const_iterator u = cluster.begin(); u != cluster.end(); u++) {
	    for (std::set<vertex_t>::const_iterator v = cluster.begin(); v != cluster.end(); v++) {
		if (*u < *v && g.is_connected(*u, *v))
		    value++;
	    }
	}
	if (verbose) {
	    fprintf(stderr, " n = %4zd m = %5d", cluster.size(), value);
	    // for (std::set<vertex_t>::const_iterator u = cluster.begin(); u != cluster.end(); u++)
	    // 	fprintf(stderr, "%d ", *u);
	    // fprintf(stderr, "\n");
	}
	if (verbose && usedILP)
	    fprintf(stderr, " %7.2f %7.2f", timerILP.total(), timerCuttingPlanes.total());
	if (verbose)
	    fprintf(stderr, "\n");

	sprintf(buf, "z_%zd", clusterVars.size());
	LP::Var zi = model.addVar(0, 1, -value, buf);
	clusterVars.push_back(zi);
	clusters.push_back(cluster);

	// add to appropriate vertex constraints
	for (std::set<vertex_t>::const_iterator u = cluster.begin(); u != cluster.end(); u++)
	    vertexConstraints[*u].setCoefficient(zi, 1.0);
    }

    if (verbose)
	fprintf(stderr, "cannot find more clusters\n");

    for (size_t i = 0; i < clusterVars.size(); i++)
	clusterVars[i].setIntegral(true);
    model.solve();

    int colors[n];
    for (size_t i = 0; i < clusters.size(); i++) {
	if (!clusterVars[i].value())
	    continue;
	for (std::set<vertex_t>::const_iterator it = clusters[i].begin(); it != clusters[i].end(); it++)
	    colors[*it] = i;
    }

    std::vector<Edge> del;
    for (size_t u = 0; u < n; u++) {
	for (NeighborIt ni = g.neighbors_begin(u); ni != g.neighbors_end(u); ++ni) {
	    size_t v = ni.n();
	    if (!(u < v))
		continue;
	    if (colors[u] != colors[v])
		del.push_back(Edge(u, v));
	}
    }

    delete pmodel;
    delete pauxModel;
    return del;
}

#else

#include <algorithm>
#include <vector>
#include <set>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#define IL_STD
#include <ilcplex/ilocplex.h>

extern "C" {
#include <caml/alloc.h>
#include <caml/callback.h>
#include <caml/memory.h>
}

#include "CGraph.hh"

#define EPS (1e-6)

class Timer {
public:
    Timer() {
	total_.tv_sec = total_.tv_nsec = 0;
    }
    void start() {
	if (clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start_)) {
	    fprintf(stderr, "error with clock_gettime(CLOCK_PROCESS_CPUTIME_ID)\n");
	    exit(1);
	}
	// struct timespec res;
	// clock_getres(CLOCK_PROCESS_CPUTIME_ID, &res);
	// fprintf(stderr, "timer resolution: %ldns\n", res.tv_nsec);
    }
    void stop() {
	struct timespec stop;
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &stop);
	total_.tv_sec  += stop.tv_sec  - start_.tv_sec;
	total_.tv_nsec += stop.tv_nsec - start_.tv_nsec;
	if (total_.tv_nsec > long(1e9)) {
	    total_.tv_nsec -= long(1e9);
	    total_.tv_sec += 1;
	} else if (total_.tv_nsec < 0) {
	    total_.tv_nsec += long(1e9);
	    total_.tv_sec -= 1;
	}
    }
    double total() {
	return total_.tv_sec + total_.tv_nsec / 1.0e9;
    }

private:
    struct timespec start_, total_;
};

ILOMIPCALLBACK0(GoodEnoughCallback) {
    if (hasIncumbent()) {
	IloNum objval = getIncumbentObjValue();
//	IloNum bound  = getBestObjValue();
//	IloNum gap = fabs(objval - bound) / (1.0 + fabs(bound));
	if (objval < -EPS)
	    this->abort();
    }
}

std::vector<Edge> hcd_colgen(const Graph& g, int /*timeLimit*/, bool verbose) {
    CAMLlocal3(v_weights, v_tuple, v_vs);
    int n = g.num_vertices();
    v_weights = caml_alloc(n, Double_array_tag);
    value* colgen_heuristic = caml_named_value("colgen_heuristic");

    Timer t_rel_master, t_aux_heur, t_aux_ilp, t_master;

    bool ILPverbose = false;
    //std::cerr << g << std::endl;

    // main problem
    std::vector<IloNumVar> clusterVars;
    std::vector<std::set<vertex_t> > clusters;
    std::vector<IloRange> vertexConstraints;

    IloEnv env;
    IloModel model(env);
    IloObjective clusterValues = IloAdd(model, IloMinimize(env));

    // initial z variables and constraints
    char buf[64];
    for (int i = 0; i < n; i++) {
	sprintf(buf, "z_%d", i);
	IloNumVar zi(env, 0.0, 1.0, buf);
	//model.add(zi);
	clusterVars.push_back(zi);
	IloRange r = zi == 1;
	model.add(r);
	vertexConstraints.push_back(r);
	std::set<vertex_t> cluster;
	cluster.insert(i);
	clusters.push_back(cluster);
	clusterValues.setLinearCoef(zi, -0.5);
    }

    // suppress annoying "license manager" output
    int backup, redirection;
    fflush(stderr);
    backup = dup(2);
    redirection = open("/dev/null", O_WRONLY);
    dup2(redirection, 2);
    close(redirection); 
    IloCplex solver(model);
    fflush(stderr);
    dup2(backup, 2);
    close(backup);
    
    if (!ILPverbose)
	solver.setOut(env.getNullStream());

    // auxiliary MIP
    IloModel auxModel(env);
    IloObjective objective = IloAdd(auxModel, IloMinimize(env));
    auxModel.add(objective);
    std::vector<IloBoolVar> auxVertexVars;
    // y_v = 1 <=> vertex v belongs to the cluster maximizing the objective
    for (int i = 0; i < n; i++) {
	sprintf(buf, "y_%d", i);
	IloBoolVar yi(env, buf);
	auxModel.add(yi);
	auxVertexVars.push_back(yi);
    }
    // cluster size constraint
    IloIntVar d(env, 0, IloIntMax, "d");
    auxModel.add(d);
    IloExpr sum(env);
    for (int i = 0; i < n; i++)
	sum += auxVertexVars[i];
    auxModel.add(sum == d);
    auxModel.add(d >= 3); // cutting plane, not needed for correctness

    // x_e: edge e belongs to the cluster maximizing the objective
    for (vertex_t u = 0; u < g.num_vertices(); u++) {
	for (NeighborIt ni = g.neighbors_begin(u); ni != g.neighbors_end(u); ++ni) {
	    vertex_t v = ni.n();
	    if (!(u < v))
		continue;
	    sprintf(buf, "x_%d_%d", u, v);
	    IloBoolVar xuv(env, buf);
	    auxModel.add(xuv);
	    objective.setLinearCoef(xuv, -1);
	    auxModel.add(xuv <= auxVertexVars[u]);
	    auxModel.add(xuv <= auxVertexVars[v]);
	}
    }

    // constraints that enforce highly connectedness
    for (vertex_t u = 0; u < n; u++) {
	IloExpr sum(env);
	for (NeighborIt ni = g.neighbors_begin(u); ni != g.neighbors_end(u); ++ni)
	    sum += auxVertexVars[ni.n()];
	auxModel.add(IloIfThen(env, auxVertexVars[u] == 1, sum - 0.5 * d >= 0.5));
	//double bigM = 0.5 * n + 0.5;
	//auxModel.add(sum - 0.5 * d + bigM * (1-auxVertexVars[u]) >= 0.5);

    }

    IloCplex auxSolver(auxModel);
    if (!ILPverbose)
	auxSolver.setOut(env.getNullStream());
    auxSolver.use(GoodEnoughCallback(env));
    auxSolver.setParam(IloCplex::MIPEmphasis, CPX_MIPEMPHASIS_FEASIBILITY);
    // suppress "Warning: Control callbacks may disable some MIP features."
    auxSolver.setWarning(env.getNullStream());

    // generate columns
    for (;;) {
	if (verbose)
	    fprintf(stderr, "timings: relaxed master %7.2fs, aux heuristic %7.2f, aux ILP %7.2f, master %7.2f\n",
		    t_rel_master.total(), t_aux_heur.total(), t_aux_ilp.total(), t_master.total());

	//solver.exportModel("/tmp/stdout.lp");

#if 0
	solver.exportModel("primal.sav");
	double objshift = 0;
	{
	    int status;
	    CPXENVptr env = CPXopenCPLEX(&status);
	    CPXLPptr lp = CPXcreateprob(env, &status, "dual");
	    CPXreadcopyprob(env, lp, "primal.sav", 0);
	    CPXdualwrite(env, lp, "dual.mps", &objshift);
	    CPXreadcopyprob(env, lp, "dual.mps", 0);
	    CPXwriteprob(env, lp, "dual.lp", 0);
	    CPXfreeprob(env, &lp);
	    CPXcloseCPLEX(&env);
	}

	static int i = 0;
	if (++i == 2)
	    exit (0);
#endif
	
	//solver.exportModel("/tmp/stdout.lp");
	if (verbose)
	    fprintf(stderr, "solving master problem, %ld columns... ", solver.getNcols());
	t_rel_master.start();
	solver.solve();
	t_rel_master.stop();
	if (verbose)
	    fprintf(stderr, "solved, objective %f\n", solver.getObjValue());

#if 0
	for (size_t i = 0; i < clusters.size(); i++) {
	    fprintf(stderr, "cluster %6.5f", solver.getValue(clusterVars[i]));
	    for (std::set<vertex_t>::const_iterator it = clusters[i].begin(); it != clusters[i].end(); it++)
		fprintf(stderr, " %d", *it);
	    fprintf(stderr, "\n");
	}
	
	for (int i = 0; i < n; i++) {
	    std::cerr << vertexConstraints[i];
	    fprintf(stderr, ": dual %d %6.5f\n", i, solver.getDual(vertexConstraints[i]));
	}
#endif
	std::set<vertex_t> cluster;

	for (int i = 0; i < n; ++i)
	    Store_double_field(v_weights, i, -solver.getDual(vertexConstraints[i]));
	t_aux_heur.start();
	v_tuple = caml_callback(*colgen_heuristic, v_weights);
	t_aux_heur.stop();
	v_vs = Field(v_tuple, 0);
	double vs_val = Double_val(Field(v_tuple, 1));
	if (vs_val < -EPS) {
	    if (verbose)
		fprintf(stderr, "column generation heuristic found cluster of value %f\n", vs_val);
	    for (size_t i = 0; i < Wosize_val(v_vs); ++i)
		cluster.insert(Int_val(Field(v_vs, i)));
       } else {
	    if (verbose)
		fprintf(stderr, "column generation heuristic found no cluster\n");
	    for (int i = 0; i < n; i++)
		objective.setLinearCoef(auxVertexVars[i], -solver.getDual(vertexConstraints[i]));

	    //auxSolver.exportModel("/tmp/stdout.lp");

	    if (verbose)
		fprintf(stderr, "solving column generation ILP, %ld columns... ", auxSolver.getNcols());
	    t_aux_ilp.start();
	    auxSolver.solve();
	    t_aux_ilp.stop();

	    if (auxSolver.getStatus() != IloAlgorithm::Optimal &&
		auxSolver.getStatus() != IloAlgorithm::Feasible &&
		auxSolver.getStatus() != IloAlgorithm::Infeasible) {
		std::cerr << "Unknown CPLEX status: " << auxSolver.getStatus() << std::endl;
		abort();
	    }
	    if (auxSolver.getStatus() == IloAlgorithm::Infeasible || auxSolver.getObjValue() >= -EPS)
		break;
	    if (verbose)
		fprintf(stderr, "solved, objective %f\n", auxSolver.getObjValue());

	    for (vertex_t u = 0; u < n; u++) {
		double x = auxSolver.getValue(auxVertexVars[u]);
		assert(fabs(x - 0.0) < EPS || fabs(x - 1.0) < EPS);
		if (x > 0.5)
		    cluster.insert(u);
	    }
	}

	// value of cluster
	int value = 0;
	for (std::set<vertex_t>::const_iterator u = cluster.begin(); u != cluster.end(); u++) {
	    for (std::set<vertex_t>::const_iterator v = cluster.begin(); v != cluster.end(); v++) {
		if (*u < *v && g.is_connected(*u, *v))
		    value++;
	    }
	}
	if (verbose) {
	    fprintf(stderr, "found cluster of size %zd and value %d: ", cluster.size(), value);
	    for (std::set<vertex_t>::const_iterator u = cluster.begin(); u != cluster.end(); u++)
		fprintf(stderr, "%d ", *u);
	    fprintf(stderr, "\n");
	}

	IloNumVar zi(env, 0.0, 1.0);
	model.add(zi);
	clusterVars.push_back(zi);
	clusters.push_back(cluster);
	// add to objective
	clusterValues.setLinearCoef(zi, -value);
	// add to appropriate vertex constraints
	for (std::set<vertex_t>::const_iterator u = cluster.begin(); u != cluster.end(); u++)
	    vertexConstraints[*u].setLinearCoef(zi, 1.0);
    }

    if (verbose)
	fprintf(stderr, "cannot find more clusters\n");
    for (size_t i = 0; i < clusterVars.size(); i++)
	model.add(IloConversion(env, clusterVars[i], ILOINT));
    //solver.exportModel("/tmp/stdout.lp");
    t_master.start();
    solver.solve();
    t_master.stop();

    if (verbose)
	fprintf(stderr, "timings: relaxed master %7.2fs, aux heuristic %7.2f, aux ILP %7.2f, master %7.2f\n",
		t_rel_master.total(), t_aux_heur.total(), t_aux_ilp.total(), t_master.total());
    
    //fprintf(stderr, "solution\n");
    int colors[n];
    for (size_t i = 0; i < clusters.size(); i++) {
	if (!solver.getValue(clusterVars[i]))
	    continue;
	for (std::set<vertex_t>::const_iterator it = clusters[i].begin(); it != clusters[i].end(); it++)
	    colors[*it] = i;
	    //fprintf(stderr, " %d", *it);
	//fprintf(stderr, "\n");
    }

    std::vector<Edge> del;
    for (vertex_t u = 0; u < n; u++) {
	IloExpr sum(env);
	for (NeighborIt ni = g.neighbors_begin(u); ni != g.neighbors_end(u); ++ni) {
	    vertex_t v = ni.n();
	    if (!(u < v))
		continue;
	    if (colors[u] != colors[v])
		del.push_back(Edge(u, v));
	}
    }

    return del;
}
#endif
