/* hcd -- solve the Highly Connected Deletion problem
   Copyright (C) 2012  Falk Hüffner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  */

#include <string.h>

#include <algorithm>
#include <limits>

#include "CGraph.hh"
#include "ILPCplex.hh"
#include "ILPGurobi.hh"

enum Solver {
    CPLEX,
    GUROBI,
    COLGEN,
};

class Problem : public ILPCallback {
public:
    Problem(const Graph& g, ILP& ilp, bool verbose);
    std::vector<Edge> solve();
    void call(CallbackILP& ilp, const double *x, bool isIntegral);
    
private:
    const Graph& g_;
    Graph gx_;
    std::unordered_map<Edge, int> edge_vars_;
    std::vector<Edge> var_edges_;
    int vertex_var(vertex_t u) const { return edge_vars_.size() + u; }
    vertex_t var_vertex(int x) const { return x - edge_vars_.size(); }
    bool verbose_;
    ILP& ilp_;
};

Problem::Problem(const Graph& g, ILP& ilp, bool verbose)
    : g_(g),
      gx_(g),
      verbose_(verbose),
      ilp_(ilp)
{
    if (verbose)
	fprintf(stderr, "building extended graph...\n");
    ilp.setVerbose(verbose);

    std::vector<double> obj;
    double obj_coeff = 0.0;
    std::vector<char*> colname;
    for (vertex_t u = 0; u < g.num_vertices(); u++) {
	for (vertex_t v = u + 1; v < g.num_vertices(); v++) {
	    if (!g.is_connected(u, v)) {
		int common = 0;
		for (NeighborIt ni = g.neighbors_begin(u); ni != g.neighbors_end(u); ++ni) {
		    vertex_t w = ni.n();
		    if (g.is_connected(v, w)) {
			if (++common >= 3)
			    break;
		    }
		}
		if (common < 3)
		    continue;
		gx_.connect(u, v, 1);
		obj.push_back(0.0);
	    } else {
		obj.push_back(-1.0);
		obj_coeff += 1;
	    }
	    edge_vars_[Edge(u, v)] = var_edges_.size();
	    var_edges_.push_back(Edge(u, v));
	    char buf[100];
	    sprintf(buf, "x_%d_%d", u, v);  
	    colname.push_back(strdup(buf));
	}
    }
    std::vector<double> values(edge_vars_.size());
    std::vector<ILPVarType> xctype(edge_vars_.size(), VAR_BIN);
    ilp_.addVariables(edge_vars_.size(), &obj[0], NULL, NULL, &xctype[0], &colname[0]);
    for (std::size_t i = 0; i < colname.size(); i++)
        free(colname[i]);
    if (verbose)
        fprintf(stderr, "added %zd edge variables\n", edge_vars_.size());

    if (1) {
	std::vector<double> obj;
	std::vector<char*> colname;
	for (vertex_t u = 0; u < g.num_vertices(); u++) {
	    obj.push_back(0.0);
	    char buf[100];
	    sprintf(buf, "x_%d", u);
	    colname.push_back(strdup(buf));
	}
	std::vector<ILPVarType> xctype(obj.size(), VAR_BIN);
	ilp_.addVariables(obj.size(), &obj[0], NULL, NULL, &xctype[0], &colname[0]);
	for (std::size_t i = 0; i < colname.size(); i++)
	  free(colname[i]);
	if (verbose)
	    fprintf(stderr, "added %zd vertex variables\n", edge_vars_.size());
    }

    int rcnt = 0;
    int nzcnt = 0;
    std::vector<int> rmatbeg;
    std::vector<int> rmatind;
    std::vector<double> rmatval;
    std::vector<double> rhs;
    std::vector<ILPConstrSense> sense;

    // add edge constraints
    // ∀ {u, v} ∈ E: e_uv ≤ u
    //               e_uv ≤ v
    for (vertex_t u = 0; u < g_.num_vertices(); u++) {
	for (NeighborIt ni = g_.neighbors_begin(u); ni != g_.neighbors_end(u); ++ni) {
	    vertex_t v = ni.n();
	    if (!(u < v))
		continue;
	    rmatbeg.push_back(nzcnt);
	    rmatval.push_back(1.0);
	    rmatind.push_back(edge_vars_[Edge(u, v)]);
	    nzcnt++;
	    rmatval.push_back(-1.0);
	    rmatind.push_back(vertex_var(u));
	    nzcnt++;
	    sense.push_back(CONSTR_LEQ);
	    rhs.push_back(0.0);
	    rcnt++;
	    
	    rmatbeg.push_back(nzcnt);
	    rmatval.push_back(1.0);
	    rmatind.push_back(edge_vars_[Edge(u, v)]);
	    nzcnt++;
	    rmatval.push_back(-1.0);
	    rmatind.push_back(vertex_var(v));
	    nzcnt++;
	    sense.push_back(CONSTR_LEQ);
	    rhs.push_back(0.0);
	    rcnt++;
	}
    }

    // add degree constraints
    // ∀ u ∈ V: ∑_(v ∈ N(u)) e_uv - ∑_(v ∈ N+(u)) e_uv - 2v_u ≥ 0
    for (vertex_t u = 0; u < g.num_vertices(); u++) {
	rmatbeg.push_back(nzcnt);
	for (NeighborIt ni = gx_.neighbors_begin(u); ni != gx_.neighbors_end(u); ++ni) {
	    vertex_t v = ni.n();
	    if (g_.is_connected(u, v))
		rmatval.push_back(+1.0);
	    else
		rmatval.push_back(-1.0);
	    rmatind.push_back(edge_vars_[Edge(u, v)]);
	    nzcnt++;
	}
	rmatval.push_back(-1.0);
	rmatind.push_back(vertex_var(u));
	nzcnt++;

	sense.push_back(CONSTR_GEQ);
	rhs.push_back(0.0);
	rcnt++;
    }

    ilp_.addConstraints(rcnt, nzcnt, &rhs[0], &sense[0], &rmatbeg[0], &rmatind[0], &rmatval[0]);

    if (verbose)
        fprintf(stderr, "added %d constraints with %d nonzero entries\n", rcnt, nzcnt);

    // add P3 constraints with two variables
    if (1) {
	int rcnt = 0;
	int nzcnt = 0;
	std::vector<int> rmatbeg;
	std::vector<int> rmatind;
	std::vector<double> rmatval;

	for (int u = 0; u < g_.num_vertices(); u++) {
	    for (int v = 0; v < g_.num_vertices(); v++) {
		if (v == u || !g_.is_connected(u, v))
		    continue;
		for (int w = 0; w < g_.num_vertices(); w++) {
		    if (w == u || w == v || !g_.is_connected(u, w) || gx_.is_connected(v, w))
			continue;
		    rmatval.push_back(1.0);
		    rmatind.push_back(edge_vars_[Edge(u, v)]);
		    rmatval.push_back(1.0);
		    rmatind.push_back(edge_vars_[Edge(u, w)]);
		    nzcnt += 2;
		    rcnt++;
		    rmatbeg.push_back(nzcnt);
		}
	    }
	}
	std::vector<double> rhs(rcnt, 1.0);
	std::vector<ILPConstrSense> sense(rcnt, CONSTR_LEQ);

	ilp_.addConstraints(rcnt, nzcnt, &rhs[0], &sense[0], &rmatbeg[0], &rmatind[0], &rmatval[0]);

	if (verbose)
	    fprintf(stderr, "added %d P3 constraints with %d nonzero entries\n", rcnt, nzcnt);
    }

    ilp_.addCallback(*this, true, true);
}

struct P3 {
    vertex_t u, v, w;
    double violation;
};

struct P3Better {
    bool operator()(const P3& a, const P3& b) const {
	return a.violation > b.violation;
    }
} p3better;


void Problem::call(CallbackILP& ilp, const double *x, bool isIntegral) {
    if (verbose_) {
        double gap = ilp.gap();
        if (!(gap < std::numeric_limits<double>::infinity()))
            fprintf(stderr, "%s callback", isIntegral ? "lazy constraints" : "user cut");
        else
            fprintf(stderr, "%s callback, gap %8.4f%%",
                    isIntegral ? "lazy constraints" : "user cut", gap * 100.0);
    }

    double eps = isIntegral ? 0.0 : 0.0001;
    std::vector<P3> p3s;
    for (vertex_t u = 0; u < gx_.num_vertices(); u++) {
	for (NeighborIt nv = gx_.neighbors_begin(u); nv != gx_.neighbors_end(u); ++nv) {
	    vertex_t v = nv.n();
	    double uv = x[edge_vars_[Edge(u, v)]];
	    for (NeighborIt nw = gx_.neighbors_begin(u); nw != gx_.neighbors_end(u); ++nw) {
		vertex_t w = nw.n();
		if (!(v < w))
		    continue;
		double uw = x[edge_vars_[Edge(u, w)]];
		double vw = gx_.is_connected(v, w) ? x[edge_vars_[Edge(v, w)]] : 0.0;
		double violation = uv + uw - vw - 1.0;
		if (violation > eps) {
		    P3 p3 = {u, v, w, violation};
		    p3s.push_back(p3);
		}
	    }
	}
    }
    if (p3s.size() == 0) {
	if (verbose_)
	    fprintf(stderr, ": no P3s found\n");
	return;
    }
    std::random_shuffle(p3s.begin(), p3s.end());
    std::sort(p3s.begin(), p3s.end(), p3better);
    const size_t MAX_P3S = 300;
    size_t num_p3s = std::min(p3s.size(), MAX_P3S);
    for (size_t i = 0; i < num_p3s; i++) {
	vertex_t u = p3s[i].u;
	vertex_t v = p3s[i].v;
	vertex_t w = p3s[i].w;
	int nzcnt = 0;
	int cutind[3];
	double cutval[3];
				    { cutind[nzcnt] = edge_vars_[Edge(u, v)]; cutval[nzcnt++] = +1.0; }
				    { cutind[nzcnt] = edge_vars_[Edge(u, w)]; cutval[nzcnt++] = +1.0; }
	if (gx_.is_connected(v, w)) { cutind[nzcnt] = edge_vars_[Edge(v, w)]; cutval[nzcnt++] = -1.0; }
	ilp.addLazyConstraint(nzcnt, 1.0, CONSTR_LEQ, cutind, cutval);
    }
    if (verbose_)
      fprintf(stderr, ": added %zd of %zd P3s, max violation %f, min violation %f\n",
	      num_p3s, p3s.size(), p3s[0].violation, p3s[num_p3s - 1].violation);
}

std::vector<Edge> Problem::solve() {
    double values[var_edges_.size() + g_.num_vertices()];
    ilp_.optimize(values);

#if 0
    for (size_t i = 0; i < var_edges_.size(); i++)
	fprintf(stderr, "%d %d: %f\n", var_edges_[i].u, var_edges_[i].v, values[i]);
    for (int u = 0; u < g_.num_vertices(); u++)
	fprintf(stderr, "%d: %f\n", u, values[vertex_var(u)]);
#endif

    std::vector<Edge> del;
    for (int u = 0; u < g_.num_vertices(); u++) {
        for (NeighborIt it = g_.neighbors_begin(u); it != g_.neighbors_end(u); ++it) {
	    vertex_t  v = it.n();
            if (!(u < v))
		continue;
	    double x = values[edge_vars_[Edge(u, v)]];
	    if (!(sim(x, 0) || sim(x, 1)))
		abort();
	    if (x < 0.5)
		del.push_back(Edge(u, v));
	}
    }
    return del;
}

static std::vector<Edge> hcd_ilp(const Graph& g, Solver solver, bool verbose) {
    ILP* ilp;
    if (solver == CPLEX)
	ilp = new ILPCplex();
    else
	ilp = new ILPGurobi();
    Problem problem(g, *ilp, verbose);
    std::vector<Edge> del = problem.solve();
    delete ilp;
    return del;
}

std::vector<Edge> hcd_colgen(const Graph& g, int timeLimit, bool verbose);

extern "C" {

#include <caml/alloc.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>


// external hcd_ilp_c : int -> (int * int) array -> solver -> bool -> (int * int) array = "c_hcd_ilp";;

value c_hcd_ilp(value v_n, value v_edges, value v_solver, value v_time_limit, value v_verbose) {
    CAMLparam5(v_n, v_edges, v_solver, v_time_limit, v_verbose);
    CAMLlocal2(v_del, v_tuple);

    fflush(stderr);

    Solver solver = Solver(Int_val(v_solver));
    bool verbose = Int_val(v_verbose);
    int n =  Int_val(v_n);
    int time_limit = Int_val(v_time_limit);
    Graph g(n);
    for (size_t i = 0; i < Wosize_val(v_edges); i++) {
	value v_tuple = Field(v_edges, i);
	int u = Int_val(Field(v_tuple, 0));
	int v = Int_val(Field(v_tuple, 1));
	g.connect(u, v, 1);
    }

    std::vector<Edge> del;
    if (solver == COLGEN)
	del = hcd_colgen(g, time_limit, verbose);
    else
	del = hcd_ilp(g, solver, verbose);

    v_del = caml_alloc(del.size(), 0);
    for (std::size_t i = 0; i < del.size(); ++i) {
	v_tuple = caml_alloc(2, 0);
	Store_field(v_tuple, 0, Val_int(del[i].u));
	Store_field(v_tuple, 1, Val_int(del[i].v));
	Store_field(v_del, i, v_tuple);
    }

    fflush(stderr);

    CAMLreturn(v_del);
}

}
