#include <cassert>
#include <stdexcept>
#include <vector>

#include <fcntl.h>
#include <unistd.h>

#include "LPCplex.hh"

namespace LP {
    CplexModel::CplexModel(Sense sense) {
	// suppress annoying "license manager" output
	int backup, redirection;
	fflush(stderr);
	backup = dup(2);
	redirection = open("/dev/null", O_WRONLY);
	dup2(redirection, 2);
	close(redirection); 
	int status;
	env_ = CPXopenCPLEX(&status);
	fflush(stderr);
	dup2(backup, 2);
	close(backup);
	if (!env_)
	    error(status, "CplexModel::CplexModel");

	lp_ = CPXcreateprob(env_, &status, "model");
	if (!lp_)
	    error(status, "CplexModel::CplexModel");

	if (sense == Sense::MAXIMIZE)
	    CPXchgobjsen(env_, lp_, CPX_MAX);
    }

    CplexModel::~CplexModel() {
	int status = CPXcloseCPLEX(&env_);
	if (status)
	    error(status, "CplexModel::~CplexModel");
    }

    void CplexModel::setVerbose(bool verbose) {
	int status;
        if ((status = CPXsetintparam(env_, CPX_PARAM_SCRIND, verbose ? CPX_ON : CPX_OFF)))
            error(status, "CplexModel::setVerbose: CPXsetintparam(CPX_PARAM_SCRIND)");
    }

    // querying

    size_t CplexModel::numVariables() const {
	return CPXgetnumcols(env_, lp_);
    }

    size_t CplexModel::numConstraints() const {
	return CPXgetnumrows(env_, lp_);
    }

    Expr CplexModel::objective() const {
	size_t n = numVariables();
	double* obj = new double[n];
	int status = CPXgetobj(env_, lp_, obj, 0, n - 1);
	if (status)
	    error(status, "CplexModel::objective: CPXgetobj");
	Expr e;
	for (size_t i = 0; i < n; i++)
	    if (obj[i])
		e += obj[i] * makeVar(i);
	delete[] obj;

	double objoffset;
	if ((status = CPXgetobjoffset(env_, lp_, &objoffset)))
	    error(status, "CplexModel::objective: CPXgetobjoffset");
	e += objoffset;

	return e;
    }

    Model::Sense CplexModel::objectiveSense() const {
	switch (CPXgetobjsen(env_, lp_)) {
	case CPX_MIN: return Sense::MINIMIZE; break;
	case CPX_MAX: return Sense::MAXIMIZE; break;
	}
	throw std::runtime_error("CplexModel::objectiveSense: unknown objective sense");
    }

    double CplexModel::objectiveValue() const {
	double objval;
	int status = CPXgetobjval(env_, lp_, &objval);
	if (status)
	    error(status, "CplexModel::objectiveValue");
	return objval;
    }

    std::string CplexModel::varName(int v) const {
	char* name[1];
	char namestore[4096];
	int surplus;
	int status = CPXgetcolname(env_, lp_, name, namestore, sizeof namestore, &surplus, v, v);
	if (status == CPXERR_NEGATIVE_SURPLUS) {
	    size_t sizeof_namestore2 = sizeof namestore + (-surplus);
	    char* namestore2 = new char[sizeof_namestore2];
	    status = CPXgetcolname(env_, lp_, name, namestore2, sizeof_namestore2, &surplus, v, v);
	    delete[] namestore2;
	}
	if (status)
	    error(status, "CplexModel::varName");
	return name[0];
    }

    ConstrExpr CplexModel::constrExpr(const Constraint& c) const {
	int row = c.id();
	int nzcnt;
	int rmatbeg[1];
	int surplus;
	int status = CPXgetrows(env_, lp_, &nzcnt, rmatbeg, nullptr, nullptr, 0, &surplus, row, row);
	//if (status != CPXERR_NEGATIVE_SURPLUS)
	//    error(status, "CplexModel::constrExpr");
	size_t n = -surplus;
	int* rmatind = new int[n];
	double* rmatval = new double[n];
	if ((status = CPXgetrows(env_, lp_, &nzcnt, rmatbeg, rmatind, rmatval, n, &surplus, row, row)))
	    error(status, "CplexModel::constrExpr: CPXgetrows");
	assert(surplus == 0);
	assert(rmatbeg[0] == 0);

	Expr e;
	for (size_t i = 0; i < n; i++)
	    e += rmatval[i] * makeVar(rmatind[i]);
	delete[] rmatind;
	delete[] rmatval;

	double rhs;
	if ((status = CPXgetrhs(env_, lp_, &rhs, row, row)))
	    error(status, "CplexModel::constrExpr: CPXgetsense");
	e -= rhs;

	char cplex_sense;
	if ((status = CPXgetsense(env_, lp_, &cplex_sense, row, row)))
	    error(status, "CplexModel::constrExpr: CPXgetsense");
	ConstrExpr::Sense sense;
	switch (cplex_sense) {
	case 'L': sense = ConstrExpr::Sense::LE; break;
	case 'E': sense = ConstrExpr::Sense::EQ; break;
	case 'G': sense = ConstrExpr::Sense::GE; break;
	default: error(0, "CplexModel::constrExpr: unknown column type"); break;
	}
	return ConstrExpr(e, sense);
    }

    // modifying

    Var CplexModel::doAddVar(double lb, double ub, double obj, const std::string& name, const char* type) {
	int v =  CPXgetnumcols(env_, lp_);
	char* names[1]{const_cast<char*>(name.c_str())};
	int status = CPXnewcols(env_, lp_, 1, &obj, &lb, &ub, type, names);
	if (status)
	    error(status, "CplexModel::addVar");
	return makeVar(v);
    }

    Var CplexModel::addVar(double lb, double ub, double obj, const std::string& name) {
	return doAddVar(lb, ub, obj, name, nullptr);
    }

    Var CplexModel::addIntVar(double lb, double ub, double obj, const std::string& name) {
	return doAddVar(lb, ub, obj, name, lb == 0.0 && ub == 1.0 ? "B" : "I");
    }

    Constraint CplexModel::addConstraint(const ConstrExpr& c) {
	int row = CPXgetnumrows(env_, lp_);
	std::vector<int> rmatind;
	std::vector<double> rmatval;
	for (auto term : c.lhs()) {
	    rmatind.push_back(term.var().id());
	    rmatval.push_back(term.coefficient());
	}
	int nzcnt = rmatind.size();
	double rhs = -c.lhs().constant();
	char sense;
	switch (c.sense()) {
	case ConstrExpr::Sense::LE: sense = 'L'; break;
	case ConstrExpr::Sense::EQ: sense = 'E'; break;
	case ConstrExpr::Sense::GE: sense = 'G'; break;
	}
	int rmatbeg[1] = {0};
	int status = CPXaddrows(env_, lp_, 0, 1, nzcnt, &rhs, &sense, rmatbeg,
				&rmatind[0], &rmatval[0], nullptr, nullptr);
	if (status)
	    error(status, "CplexModel::addConstraint");
	return makeConstraint(row);
    }

    Constraint CplexModel::addConditionalConstraint(bool positive, const Var& v, double /*bigM*/,
						    const ConstrExpr& c) {
	int row = CPXgetnumrows(env_, lp_);
	std::vector<int> rmatind;
	std::vector<double> rmatval;
	for (auto term : c.lhs()) {
	    rmatind.push_back(term.var().id());
	    rmatval.push_back(term.coefficient());
	}
	int nzcnt = rmatind.size();
	double rhs = -c.lhs().constant();
	char sense;
	switch (c.sense()) {
	case ConstrExpr::Sense::LE: sense = 'L'; break;
	case ConstrExpr::Sense::EQ: sense = 'E'; break;
	case ConstrExpr::Sense::GE: sense = 'G'; break;
	default: abort();
	}
	int status = CPXaddindconstr(env_, lp_, v.id(), !positive, nzcnt, rhs, sense,
				     &rmatind[0], &rmatval[0], nullptr);
	if (status)
	    error(status, "CplexModel::addConstraint");
	return makeConstraint(row);
    }

    void CplexModel::solve() {
	minimizing_ = objectiveSense() == Sense::MINIMIZE; // for GoodEnough
	int status;
	if (CPXgetnumint(env_, lp_) || CPXgetnumbin(env_, lp_))
	    status = CPXmipopt(env_, lp_);
	else
	    status = CPXlpopt(env_, lp_);
	if (status)
	    error(status, "CplexModel::solve");
    }

    Status CplexModel::status() const {
	int status = CPXgetstat(env_, lp_);
	switch (status) {
	case CPX_STAT_OPTIMAL:
	case CPXMIP_OPTIMAL:
	case CPXMIP_OPTIMAL_TOL:
	    return Status::OPTIMAL; break;
	case CPXMIP_INFEASIBLE:
	    return Status::INFEASIBLE; break;
	case CPXMIP_INForUNBD:
	    return Status::INFEASIBLE_OR_UNBOUNDED; break;
	case CPXMIP_ABORT_FEAS:
	    return Status::ABORTED_FEASIBLE; break;
	default:
	    printf("unknown CPLEX status = %d\n", status);
	    exit(1);
	    break;
	}
    }

    bool CplexModel::varIsIntegral(const Var& v) const {
	char xctype;
	int status = CPXgetctype(env_, lp_, &xctype, v.id(), v.id());
	if (status == CPXERR_NOT_MIP)
	    return false;
	if (status)
	    error(status, "CplexModel::varIsIntegral: CPXgetctype");
	switch (xctype) {
	case 'I': case 'B': return true;  break;
	case 'C':           return false; break;
	default:
	    error(status, "CplexModel::varIsIntegral: unknown column type " + xctype);
	    break;
	}
	return false;		// suppress warning
    }

    double CplexModel::varlb(const Var& v) const {
	double lb;
	int status = CPXgetlb(env_, lp_, &lb, v.id(), v.id());
	if (status)
	    error(status, "CplexModel::varlbl");
	return lb;
    }

    double CplexModel::varub(const Var& v) const {
	double ub;
	int status = CPXgetub(env_, lp_, &ub, v.id(), v.id());
	if (status)
	    error(status, "CplexModel::varub");
	return ub;
    }

    double CplexModel::varValue(const Var& v) const {
	double x;
	int status = CPXgetx(env_, lp_, &x, v.id(), v.id());
	if (status)
	    error(status, "CplexModel::varValue");
	return x;
    }

    double CplexModel::dualValue(const Constraint& c) const {
	double x;
	int status = CPXgetpi(env_, lp_, &x, c.id(), c.id());
	if (status)
	    error(status, "CplexModel::dualValue");
	return x;
    }

    void CplexModel::varChangeBounds(const Var& v, double lb, double ub) const {
	const int count = 1;
	int indices[count] = { v.id() };
	double lbs[count] = { lb };
	double ubs[count] = { ub };
	int status;

	if ((status = CPXchgbds(env_, lp_, count, indices, "L", lbs)))
	    error(status, "CplexModel::varChangeBounds(L)");
	if ((status = CPXchgbds(env_, lp_, count, indices, "U", ubs)))
	    error(status, "CplexModel::varChangeBounds(U)");
    }

    void CplexModel::varSetIntegral(const Var& v, bool integral) const {
	const int indices[1] = { v.id() };
	const char xctype[1] = { integral ? CPX_INTEGER : CPX_CONTINUOUS };
	int status = CPXchgctype(env_, lp_, 1, indices, xctype);
	if (status)
	    error(status, "CplexModel::varSetIntegral: CPXchgctype");
    }

    void CplexModel::setCoefficient(const Constraint& c, const Var& v, double x) const {
	int status = CPXchgcoef(env_, lp_, c.id(), v.id(), x);
	if (status)
	    error(status, "CplexModel::setCoefficient: CPXchgcoef");
    }

    int goodEnoughCallback(CPXCENVptr env, void *cbdata, int wherefrom, void* cbhandle) {
	const CplexModel* model = static_cast<const CplexModel*>(cbhandle);	
	int status;
	int haveIncumbent;
	if ((status = CPXgetcallbackinfo(env, cbdata, wherefrom,
					 CPX_CALLBACK_INFO_MIP_FEAS, &haveIncumbent))) {
	    char buffer[CPXMESSAGEBUFSIZE];
	    throw std::runtime_error(std::string("CPXgetcallbackinfo: ") + CPXgeterrorstring(env, status, buffer));
	}
	if (!haveIncumbent)
	    return 0;
	double objval;
	if ((status = CPXgetcallbackinfo(env, cbdata, wherefrom,
					 CPX_CALLBACK_INFO_BEST_INTEGER, &objval))) {
	    char buffer[CPXMESSAGEBUFSIZE];
	    throw std::runtime_error(std::string("CPXgetcallbackinfo: ") + CPXgeterrorstring(env, status, buffer));
	}
	if (model->minimizing_) {
	    if (objval <= model->goodEnough())
		return 1;	// nonzero status causes abort
	} else {
	    if (objval >= model->goodEnough())
		return 1;
	}
	return 0;
    }
    
    void CplexModel::setObjectiveCoeff(const Var& v, double c) {
	int indices[1] = { v.id() };
	double values[1] = { c };
	int status = CPXchgobj(env_, lp_, 1, indices, values);
	if (status)
	    error(status, "CplexModel::setObjectiveCoeff");
    }

    void CplexModel::setObjectiveSense(Sense s) {
#if CPX_VERSION_VERSION > 12 || (CPX_VERSION_VERSION == 12 && CPX_VERSION_RELEASE >= 5)
	int status = CPXchgobjsen(env_, lp_, s == Sense::MINIMIZE ? CPX_MIN : CPX_MAX);
	if (status)
	    error(status, "CplexModel::setObjectiveSense");
#else
	CPXchgobjsen(env_, lp_, s == Sense::MINIMIZE ? CPX_MIN : CPX_MAX);
#endif
    }
    
    void CplexModel::setGoodEnough(double x) {
	goodEnough_ = x;
	int status = CPXsetinfocallbackfunc(env_, goodEnoughCallback, this);
	if (status)
	    error(status, "CplexModel::setGoodEnough");
    }

    void CplexModel::setEmphasis(Emphasis e) {
	int emph = CPX_MIPEMPHASIS_BALANCED;
	switch (e) {
	case Emphasis::BALANCED:         emph = CPX_MIPEMPHASIS_BALANCED;    break;
	case Emphasis::FIND_FEASIBLE:    emph = CPX_MIPEMPHASIS_FEASIBILITY; break;
	case Emphasis::PROVE_OPTIMALITY: emph = CPX_MIPEMPHASIS_OPTIMALITY;  break;
	case Emphasis::MOVE_BOUND:       emph = CPX_MIPEMPHASIS_BESTBOUND;   break;
	}
	int status = CPXsetintparam(env_, CPX_PARAM_MIPEMPHASIS, emph);
	if (status)
	    error(status, "CplexModel::setEmphasis");
    }

    static int cplexCallback(CPXCENVptr env, void* cbdata, int wherefrom, void* cbhandle, int* useraction_p) {
	Callback& callback = *static_cast<Callback*>(cbhandle);
	CplexCallbackModel model(env, cbdata, wherefrom);

	callback.call(model);

	*useraction_p = CPX_CALLBACK_DEFAULT;
	return 0;
    }

    void CplexModel::setCutCallback(Callback& callback, bool cuts, bool lazyConstraints) {
	if (!cuts && !lazyConstraints)
	    return;
	int status;
	// the callback wants to use variable indices of the original and not the presolved ILP
	if ((status = CPXsetintparam(env_, CPX_PARAM_MIPCBREDLP, CPX_OFF)))
	    error(status, "CPXsetintparam(CPX_PARAM_MIPCBREDLP)");
	if (cuts)
	    if ((status = CPXsetusercutcallbackfunc(env_, cplexCallback, &callback)))
		error(status, "CPXsetusercutcallbackfunc");
	if (lazyConstraints)
	    if ((status = CPXsetlazyconstraintcallbackfunc(env_, cplexCallback, &callback)))
		error(status, "CPXsetlazyconstraintcallbackfunc");
    }

    static void cplexError(CPXCENVptr env, int status, const std::string& msg) {
	char buffer[CPXMESSAGEBUFSIZE];
	CPXgeterrorstring(env, status, buffer);
	throw std::runtime_error(msg + std::string(": ") + std::string(buffer));
    }

    void CplexModel::error(int status, const std::string& msg) const {
	cplexError(env_, status, msg);
    }

    double CplexCallbackModel::gap() const {
	double gap;
	int status = CPXgetcallbackinfo(env_, cbdata_, wherefrom_, CPX_CALLBACK_INFO_MIP_REL_GAP, &gap);
	if (status)
	    cplexError(env_, status, "CplexCallbackModel::gap: CPX_CALLBACK_INFO_MIP_REL_GAP");
	return gap >= CPX_INFBOUND ? std::numeric_limits<double>::infinity() : gap;
    }

    double CplexCallbackModel::value(const Var& v) const {
	int status;
	double x;
	if ((status = CPXgetcallbacknodex(env_, cbdata_, wherefrom_, &x, v.id(), v.id())))
	    cplexError(env_, status, "CplexCallbackModel::value: CPXgetcallbacknodex");
	return x;
    }

    void CplexCallbackModel::addLazyConstraint(const ConstrExpr& c) {
	std::vector<int> indices;
	std::vector<double> coeffs;
	for (auto term : c.lhs()) {
	    indices.push_back(term.var().id());
	    coeffs.push_back(term.coefficient());
	}
	int numNonzero = indices.size();
	double rhs = -c.lhs().constant();
	char sense;
	switch (c.sense()) {
	case ConstrExpr::Sense::LE: sense = 'L'; break;
	case ConstrExpr::Sense::EQ: sense = 'E'; break;
	case ConstrExpr::Sense::GE: sense = 'G'; break;
	default: abort();
	}

	int status = CPXcutcallbackadd(env_, cbdata_, wherefrom_, numNonzero, rhs, sense,
				       &indices[0], &coeffs[0], CPX_USECUT_FORCE);
	if (status)
	    cplexError(env_, status, "CplexCallbackModel::addLazyConstraint: CPXcutcallbackadd");
    }

    void CplexCallbackModel::addUserCut(const ConstrExpr& c) {
	std::vector<int> indices;
	std::vector<double> coeffs;
	for (auto term : c.lhs()) {
	    indices.push_back(term.var().id());
	    coeffs.push_back(term.coefficient());
	}
	int numNonzero = indices.size();
	double rhs = -c.lhs().constant();
	char sense;
	switch (c.sense()) {
	case ConstrExpr::Sense::LE: sense = 'L'; break;
	case ConstrExpr::Sense::EQ: sense = 'E'; break;
	case ConstrExpr::Sense::GE: sense = 'G'; break;
	default: abort();
	}

	int status = CPXcutcallbackadd(env_, cbdata_, wherefrom_, numNonzero, rhs, sense,
				       &indices[0], &coeffs[0], CPX_USECUT_FORCE);
	if (status)
	    cplexError(env_, status, "CplexCallbackModel::addUserCut: CPXcutcallbackadd");
    }
}
