#include <cassert>

#include <vector>
#include <sstream>

#include "LPGurobi.hh"

template<typename T>
static std::string itoa(T x) {
    std::ostringstream s;
    s << x;
    return s.str();
}

namespace LP {
    GurobiModel::GurobiModel(Sense sense) : addedVars_(0), addedConstrs_(0) {
	if (!env_)
	    if (GRBloadenv(&env_, nullptr))
		error("GurobiModel::GurobiModel: GRBloadenv");
	if (GRBnewmodel(env_, &lp_, nullptr, 0, nullptr, nullptr, nullptr, nullptr, nullptr))
	    error("GurobiModel::GurobiModel: GRBnewmodel");
	if (sense == Sense::MAXIMIZE)
	    setObjectiveSense(Sense::MAXIMIZE);
	setVerbose(false);
    }

    GurobiModel::~GurobiModel() {
	if (GRBfreemodel(lp_))
	    error("GurobiModel::~GurobiModel: GRBfreemodel");
    }

    void GurobiModel::setVerbose(bool verbose) {
	if (GRBsetintparam(env(), "OutputFlag", verbose))
	    error("GurobiModel::setVerbose: GRBsetintparam(OutputFlag)");
    }

    // querying

    size_t GurobiModel::numVariables() const {
	int numVars;
	if (GRBgetintattr(lp_, "NumVars", &numVars))
	    error("GurobiModel::numVariables: GRBgetintattr(NumVars)");
	return numVars;
    }

    size_t GurobiModel::numConstraints() const {
	int numConstrs;
	if (GRBgetintattr(lp_, "NumConstrs", &numConstrs))
	    error("GurobiModel::numConstraints: GRBgetintattr(NumConstrs)");
	return numConstrs;
    }

    Expr GurobiModel::objective() const {
	update();
	size_t n = numVariables();
	double* obj = new double[n];
	if (GRBgetdblattrarray(lp_, "Obj", 0, n, obj))
	    error("GurobiModel::objective: GRBgetdblattrarray(Obj)");

	Expr e;
	for (size_t i = 0; i < n; i++)
	    if (obj[i])
		e += obj[i] * makeVar(i);
	delete[] obj;

	double objoffset;
	if (GRBgetdblattr(lp_, "ObjCon", &objoffset))
	    error("GurobiModel::objective: GRBgetdblattr(ObjCon)");
	e += objoffset;

	return e;
    }

    Model::Sense GurobiModel::objectiveSense() const {
	int sense;
	if (GRBgetintattr(lp_, "ModelSense", &sense))
	    error("GurobiModel::objectiveSense: GRBgetintattr(ModelSense)");

	switch (sense) {
	case +1: return Sense::MINIMIZE; break;
	case -1: return Sense::MAXIMIZE; break;
	}
	error("GurobiModel::objectiveSense: unknown objective sense");
	return Sense::MINIMIZE; // suppress warning
    }

    double GurobiModel::objectiveValue() const {
	double objval;
	if (GRBgetdblattr(lp_, "ObjVal", &objval))
	    error("GurobiModel::varName: GRBgetdblattr(ObjVal)");
	return objval;
    }

    std::string GurobiModel::varName(int v) const {
	char *name;
	if (GRBgetstrattrelement(lp_, "VarName", v, &name))
	    error("GurobiModel::varName: GRBgetstrattrelement(VarName)");
	return name;
    }

    ConstrExpr GurobiModel::constrExpr(const Constraint& c) const {
	int row = c.id();
	int nzcnt;
	int rmatbeg[1];
	if (GRBgetconstrs(lp_,  &nzcnt, nullptr, nullptr, nullptr, row, 1))
	    error("GurobiModel::constrExpr: GRBgetconstrs");
	int* rmatind = new int[nzcnt];
	double* rmatval = new double[nzcnt];
	if (GRBgetconstrs(lp_, &nzcnt, rmatbeg, rmatind, rmatval, row, 1))
	    error("GurobiModel::constrExpr: GRBgetconstrs");
	assert(rmatbeg[0] == 0);

	Expr e;
	for (size_t i = 0; i < size_t(nzcnt); i++)
	    e += rmatval[i] * makeVar(rmatind[i]);
	delete[] rmatind;
	delete[] rmatval;

	double rhs;
	if (GRBgetdblattrelement(lp_, "RHS", row, &rhs))
	    error("GurobiModel::constrExpr: GRBgetdblattrelement(RHS)");
	e -= rhs;

	char gurobiSense;
	if (GRBgetcharattrelement(lp_, "Sense", row, &gurobiSense))
	    error("GurobiModel::constrExpr: GRBgetcharattrelement(Sense)");
	ConstrExpr::Sense sense;
	switch (gurobiSense) {
	case '<': sense = ConstrExpr::Sense::LE; break;
	case '=': sense = ConstrExpr::Sense::EQ; break;
	case '>': sense = ConstrExpr::Sense::GE; break;
	default: error("GurobiModel::constrExpr: unknown column type"); break;
	}
	return ConstrExpr(e, sense);
    }

    // modifying

    Var GurobiModel::doAddVar(double lb, double ub, double obj, const std::string& name, char type) {
	int v = numVariables() + addedVars_++;
	if (GRBaddvar(lp_, 0, nullptr, nullptr, obj, lb, ub, type, const_cast<char*>(name.c_str())))
	    error("GurobiModel::doAddVar: GRBaddvar");
	return makeVar(v);
    }

    Var GurobiModel::addVar(double lb, double ub, double obj, const std::string& name) {
	return doAddVar(lb, ub, obj, name,  GRB_CONTINUOUS);
    }

    Var GurobiModel::addIntVar(double lb, double ub, double obj, const std::string& name) {
	return doAddVar(lb, ub, obj, name, lb == 0.0 && ub == 1.0 ? GRB_BINARY : GRB_INTEGER);
    }

    Constraint GurobiModel::addConstraint(const ConstrExpr& c) {
	if (addedVars_)
	    update();
	int row = numConstraints() + addedConstrs_++;
	std::vector<int> rmatind;
	std::vector<double> rmatval;
	for (auto term : c.lhs()) {
	    rmatind.push_back(term.var().id());
	    rmatval.push_back(term.coefficient());
	}
	int nzcnt = rmatind.size();
	double rhs = -c.lhs().constant();
	char sense = GRB_EQUAL;
	switch (c.sense()) {
	case ConstrExpr::Sense::LE: sense = GRB_LESS_EQUAL;    break;
	case ConstrExpr::Sense::EQ: sense = GRB_EQUAL;         break;
	case ConstrExpr::Sense::GE: sense = GRB_GREATER_EQUAL; break;
	}
	if (GRBaddconstr(lp_, nzcnt, &rmatind[0], &rmatval[0], sense, rhs, nullptr))
	    error("GurobiModel::addConstraint: GRBaddconstr");
	return makeConstraint(row);
    }

    Constraint GurobiModel::addConditionalConstraint(bool positive, const Var& v, double bigM,
						    const ConstrExpr& c) {
	
	// Gurobi does not support indicator constraints, we need to use the big M
	switch (c.sense()) {
	case ConstrExpr::Sense::LE:
	    return addConstraint(c.lhs() - bigM * (positive ? 1 - v : v) <= 0);
	    break;
	case ConstrExpr::Sense::GE:
	    return addConstraint(c.lhs() + bigM * (positive ? 1 - v : v) >= 0);
	    break;
	default:
	    error("GurobiModel::addConditionalConstraint: bad constraint sense");
	    abort();
	}
    }

    void GurobiModel::solve() {
	if (GRBoptimize(lp_))
	    error("GurobiModel::solve: GRBoptimize");
    }

    Status GurobiModel::status() const {
	int state;
	if (GRBgetintattr(lp_, "Status", &state))
	    error("GurobiModel::status: GRBgetintattr(Status)");
	switch (state) {
	case GRB_OPTIMAL:     return Status::OPTIMAL;                 break;
	case GRB_INF_OR_UNBD: return Status::INFEASIBLE_OR_UNBOUNDED; break;
	case GRB_INTERRUPTED: return Status::ABORTED_FEASIBLE;        break;
	default:
	    throw std::runtime_error("GurobiModel::status: unknown Gurobi status " + itoa(state));
	    break;
	}
    }

    bool GurobiModel::varIsIntegral(const Var& v) const {
	char type = varCharAttr(v, "VType");
	switch (type) {
	case GRB_INTEGER: case GRB_BINARY: return true;  break;
	case GRB_CONTINUOUS:               return false; break;
	default:
	    throw std::runtime_error("GurobiModel::status: unknown Gurobi variable type " + type);
	    break;
	}
    }

    double GurobiModel::varlb(const Var& v) const {
	return varDoubleAttr(v, "LB");
    }

    double GurobiModel::varub(const Var& v) const {
	return varDoubleAttr(v, "UB");
    }

    double GurobiModel::varValue(const Var& v) const {
	return varDoubleAttr(v, "X");
    }

    double GurobiModel::dualValue(const Constraint& c) const {
	double x;
	if (GRBgetdblattrelement(lp_, "Pi", c.id(), &x))
	    error("GurobiModel::dualValue: GRBgetdblattrelement(RC)");
	return x;
    }

    void GurobiModel::varChangeBounds(const Var& v, double lb, double ub) const {
	if (GRBsetdblattrelement(lp_, "LB", v.id(), lb))
	    error("GurobiModel::varChangeBounds: GRBsetdblattrelement(LB)");
	if (GRBsetdblattrelement(lp_, "UB", v.id(), ub))
	    error("GurobiModel::varChangeBounds: GRBsetdblattrelement(UB)");
    }

    void GurobiModel::varSetIntegral(const Var& v, bool integral) const {
	if (GRBsetcharattrelement(lp_, "VType", v.id(), integral ? 'I' : 'B'))
	    error("GurobiModel::varSetIntegral: GRBsetcharattrelement(VType)");
    }

    void GurobiModel::setCoefficient(const Constraint& c, const Var& v, double x) const {
	if (addedVars_ || addedConstrs_)
	    update();
	int cind[] = { c.id() };
	int vind[] = { v.id() };
	double val[] = { x };
	if (GRBchgcoeffs(lp_, 1, cind, vind, val))
	    error("GurobiModel::setCoefficient: GRBchgcoeffs");
    }

    void GurobiModel::setObjectiveCoeff(const Var& v, double c) {
	if (GRBsetdblattrelement(lp_, "Obj", v.id(), c))
	    error("GurobiModel::setObjectiveCoeff: GRBsetdblattrelement(Obj)");
    }

    void GurobiModel::setObjectiveSense(Sense s) {
	if (GRBsetintattr(lp_, "ModelSense", s == Sense::MINIMIZE ? +1 : -1))
	    error("GurobiModel::setObjectiveSense: GRBsetintattrelement(ModelSense)");
    }

    static int goodEnoughCallback(GRBmodel* lp, void* cbdata, int where, void* usrdata) {
	if (where != GRB_CB_MIPSOL)
	    return 0;
	int sense;
	if (GRBgetintattr(lp, "ModelSense", &sense))
	    throw std::runtime_error("GurobiModel::objectiveSense: GRBgetintattr(ModelSense)");
	double objval;
	if (GRBcbget(cbdata, where, GRB_CB_MIPSOL_OBJBST, &objval))
	    throw std::runtime_error("GRBcbget");
	const GurobiModel* model = static_cast<const GurobiModel*>(usrdata);
	if (sense == GRB_MINIMIZE) {
	    if (objval <= model->goodEnough())
		GRBterminate(lp);
	} else {
	    if (objval >= model->goodEnough())
		GRBterminate(lp);
	}
	return 0;
    }

    void GurobiModel::setGoodEnough(double x) {
	goodEnough_ = x;
	if (GRBsetcallbackfunc (lp_, goodEnoughCallback, this))
	    error("GurobiModel::setGoodEnough");
    }

    void GurobiModel::setEmphasis(Emphasis e) {
	int emph = GRB_MIPFOCUS_BALANCED;
	switch (e) {
	case Emphasis::BALANCED:         emph = GRB_MIPFOCUS_BALANCED;    break;
	case Emphasis::FIND_FEASIBLE:    emph = GRB_MIPFOCUS_FEASIBILITY; break;
	case Emphasis::PROVE_OPTIMALITY: emph = GRB_MIPFOCUS_OPTIMALITY;  break;
	case Emphasis::MOVE_BOUND:       emph = GRB_MIPFOCUS_BESTBOUND;   break;
	}
	if (GRBsetintparam(env_, GRB_INT_PAR_MIPFOCUS, emph))
	    error("GurobiModel::setEmphasis");
    }

    void GurobiModel::setCutCallback(Callback& /*callback*/, bool /*cuts*/, bool /*lazyConstraints*/) {
	
    }

    int GurobiModel::varIntAttr(const Var& v, const char* attr) const {
	int x;
	if (GRBgetintattrelement(lp_, attr, v.id(), &x))
	    error("GRBgetintattr(" + std::string(attr) + ")");
	return x;
    }

    char GurobiModel::varCharAttr(const Var& v, const char* attr) const {
	char x;
	if (GRBgetcharattrelement(lp_, attr, v.id(), &x))
	    error("GRBgetcharattr(" + std::string(attr) + ")");
	return x;
    }

    double GurobiModel::varDoubleAttr(const Var& v, const char* attr) const {
	double x;
	if (GRBgetdblattrelement(lp_, attr, v.id(), &x))
	    error("GRBgetdblattr(" + std::string(attr) + ")");
	return x;
    }

    const char* GurobiModel::varStringAttr(const Var& v, const char* attr) const {
	char *x;
	if (GRBgetstrattrelement(lp_, attr, v.id(), &x))
	    error("GRBgetdblattr(" + std::string(attr) + ")");
	return x;
    }
    
    void GurobiModel::error(const std::string& msg) const {
	throw std::runtime_error(msg + ": " + GRBgeterrormsg(env()) + "!!!");
    }

    GRBenv* GurobiModel::env_;
}
